import os
import copy
from collections import OrderedDict

import numpy as np
import yaml

from ...Workflow import Workflow 
from ...IO.BL15.ReadTimeSeries import ReadTimeSeries 
from ...INTEGRATION.Integrate import Integrate 
from ....operations.DAQ.BL15.SetFlowReactor import SetFlowReactor
from ....operations.DAQ.BL15.ReadFlowReactor import ReadFlowReactor
from ....operations.IO.YAML.SaveXRSDSystem import SaveXRSDSystem
from ....operations.DAQ.BL15.MarCCD_SISExpose import MarCCD_SISExpose
from ....operations.IO.FILESYSTEM.SSHCopy import SSHCopy
from ....operations.PROCESSING.BACKGROUND.BgSubtract import BgSubtract 
from ....operations.PROCESSING.FITTING.XRSDFit import XRSDFit 
from ....operations.PACKAGING.PIF.FlowSynthesisPIF import FlowSynthesisPIF
from ....operations.IO.CITRINATION.UploadPIF import UploadPIF

inputs = OrderedDict(
    flow_reactor = None,        #
    spec_infoclient = None,     #
    ssh_client = None,          #
    source_wavelength = None,   #
    experiment_id = 'test',     #
    bg_recipe = {},             #
    rxn_recipe = {},            #
    n_bg_exposures = 1,         #
    n_rxn_exposures = 1,        #
    delay_volume = 0.,          #
    delay_time = 0.,            #
    exposure_time = 10.,        #
    bg_header_dir = None,       #
    bg_image_dir = None,        #
    rxn_header_dir = None,      #
    rxn_image_dir = None,       #
    stop_reactor = True         #
    )

outputs = OrderedDict()

class RunFlowReactor(Workflow):
    
    def __init__(self):
        super(RunFlowReactor,self).__init__(inputs,outputs)
        self.add_operations(
            set_recipe = SetFlowReactor(),          #
            read_flow_reactor = ReadFlowReactor(),  #
            expose_mar = MarCCD_SISExpose(),        #
            collect_file = SSHCopy()                #
            )

    def run(self):
        freac = self.inputs['flow_reactor']
        rxn_id = self.inputs['experiment_id']
        src_wl = self.inputs['source_wavelength']
        exp_tm = self.inputs['exposure_time']
        # SET BACKGROUND RECIPE, DELAY
        self.operations['set_recipe'].run_with(
            flow_reactor = freac,
            recipe = self.inputs['bg_recipe'],
            delay_time = self.inputs['delay_time'],
            delay_volume = self.inputs['delay_volume']
            )
        # EXPOSE BACKGROUND IMAGES, READ FLOWREACTOR, SAVE HEADER DATA
        nbgx = self.inputs['n_bg_exposures']
        for ibg in range(nbgx):
            bg_fn_root = rxn_id+'_bg{}'.format(ibg) 
            self.operations['expose_mar'].run_with(
                spec_infoclient = self.inputs['spec_infoclient'],
                exposure_time = exp_tm,
                filename = bg_fn_root 
                )
            fr_stat = self.operations['read_flow_reactor'].run_with(flow_reactor=freac)
            bg_header_file = os.path.join(self.inputs['bg_header_dir'],bg_fn_root+'.yml')
            bg_hdr = OrderedDict(exposure_time=exp_tm,source_wavelength=src_wl) 
            bg_hdr.update(self.inputs['bg_recipe']),
            bg_hdr.update(fr_stat['header'])
            yaml.dump(bg_hdr,open(bg_header_file,'w'))
        # SET REACTION RECIPE, DELAY
        self.operations['set_recipe'].run_with(
            flow_reactor = freac,
            recipe = self.inputs['rxn_recipe'],
            delay_time = self.inputs['delay_time'],
            delay_volume = self.inputs['delay_volume']
            )
        # EXPOSE REACTION IMAGES, READ FLOWREACTOR, SAVE HEADER DATA
        nrxnx = self.inputs['n_rxn_exposures']
        for irx in range(nrxnx):
            rxn_fn_root = rxn_id+'_{}'.format(irx) 
            self.operations['expose_mar'].run_with(
                spec_infoclient = self.inputs['spec_infoclient'],
                exposure_time = exp_tm,
                filename = rxn_fn_root 
                )
            fr_stat = self.operations['read_flow_reactor'].run_with(flow_reactor=freac)
            rxn_header_file = os.path.join(self.inputs['rxn_header_dir'],rxn_fn_root+'.yml')
            rxn_hdr = OrderedDict(exposure_time=exp_tm,source_wavelength=src_wl)
            rxn_hdr.update(self.inputs['rxn_recipe']),
            rxn_hdr.update(fr_stat['header'])
            yaml.dump(rxn_hdr,open(rxn_header_file,'w'))
        # COLLECT BACKGROUND IMAGES
        for ibg in range(nbgx):
            fn = rxn_id+'_bg{}_0001.tif'.format(ibg)
            mar_path = '/home/data/'+fn
            local_path = os.path.join(self.inputs['bg_image_dir'],fn)
            self.operations['collect_file'].run_with(
                ssh_client = self.inputs['ssh_client'],
                remote_path = mar_path,
                local_path = local_path
                ) 
        # COLLECT REACTION IMAGES
        for irx in range(nrxnx):
            fn = rxn_id+'_{}_0001.tif'.format(irx)
            mar_path = '/home/data/'+fn
            local_path = os.path.join(self.inputs['rxn_image_dir'],fn)
            self.operations['collect_file'].run_with(
                ssh_client = self.inputs['ssh_client'],
                remote_path = mar_path,
                local_path = local_path
                )

        # STOP THE FLOW REACTOR
        if self.inputs['stop_reactor']: freac.stop()
        return self.outputs




from collections import OrderedDict
import glob
import os
import re

from ...Operation import Operation

inputs=OrderedDict(
    dir_path=None,
    regex='*',
    filter_regex=None)
outputs=OrderedDict(
    file_list=None,
    filename_list=None)

class BuildFileList(Operation):
    """
    Read a directory and filter its contents with a regular expression
    to form a list of file paths.
    """

    def __init__(self):
        super(BuildFileList,self).__init__(inputs,outputs)
        self.input_doc['dir_path'] = 'path to directory containing files'
        self.input_doc['regex'] = 'unix-like regex to select files'
        self.input_doc['filter_regex'] = 'regex used to filter output'
        
    def run(self):
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        frx = self.inputs['filter_regex']
        globex = os.path.join(dirpath,rx)
        self.message_callback('seeking files matching {}'.format(globex))
        fl = glob.glob(globex)
        if frx is not None:
            fl = list(filter(re.compile(frx).match,fl))
        fnamel = [os.path.split(p)[-1] for p in fl] 
        self.outputs['file_list'] = fl
        self.outputs['filename_list'] = fnamel




"""
The INPUT.CALIBRATION category has operations
for reading in calibration parameters
and converting them between different formats.
Some of the common formats are described here.
Over time, these descriptions should improve.
Contact the paws developers to contribute information 
or report inconsistencies. 


PONI (PyFAI) FORMAT
-------------------

PONI: point of normal incidence.
This is the format used internally by the PyFAI 
(Python Fast Azimuthal Integration) package.
PONI format projects the point-shaped sample orthogonally onto projector plane,
and gives the coordinates of that projection as the PONI,
such that the sample to PONI distance is the shortest distance from sample to detector plane.
coordinate axes: x1 vertical, x2 and x3 horizontal, x3 along beam.
detector axes: with zero rotations, d1 vertical, d2 horizontal, d3 along beam.
axes defined on C format, first dimension is vertical, second dimension is horizontal.
the first dimension (vertical) is fast, the second dimension (horizontal) is slow. 

PONI dict keys and definitions:
- 'dist': distance in meters from sample to PONI on detector plane
- 'poni1': vertical coordinate of PONI on detector axes, in meters
- 'poni2': horizontal coordinate of PONI on detector axes, in meters
- 'rot1': rotation of detector body about x1, applied first, radians
- 'rot2': rotation of detector body about x2, applied second, radians
- 'rot3': rotation of detector body about beam axis x3, applied third, radians
- 'pixel1': pixel dimension along d1 (vertical), meters
- 'pixel2': pixel dimension along d2 (horizontal), meters
- 'wavelength': wavelength in meters
- 'fpolz': polarization factor- not actually a PONI parameter, but it's ok to put it here 
- 'detector': optional pyFAI detector object
- 'splineFile' optional spline file describing detector distortion


NIKA FORMAT
-----------

The calibration performed by the Nika software package
uses a calibrant image, the rectangular pixel dimensions (in mm), 
and the wavelength (in Angstrom), 
to solve the sample to CCD distance in mm,
the position at which the beam axis intersects the detector plane in pixels,
and the horizontal and vertical tilts of the detector in degrees.

Nika does not generate a file to save calibration parameters,
so they have to be recorded by hand in a file.
Paws Operations should be written to read them from a file 
in the following format (one parameter=value per line, no spaces):
- sample_to_CCD_mm=____
- pixel_size_x_mm=____
- pixel_size_y_mm=____
- beam_center_x_pix=____
- beam_center_y_pix=____
- horizontal_tilt_deg=____
- vertical_tilt_deg=____ 
- wavelength_A=____ 


FIT2D FORMAT
------------

Detector plane origin is the bottom left corner of the detector.

Fit2D dict keys and definitions:
- 'directDist': direct distance to detector plane along beam axis, in mm
- 'centerX': horizontal position on the detector plane where the beam intersects, in px
- 'centerY': vertical position on the detector plan where the beam intersects, in px
- 'pixelX': horizontal size of pixel, in um 
- 'pixelY': vertical size of pixel, in um 
- 'tilt': detector tilt in degrees (TODO:clarify)
- 'tiltPlanRotation': detector rotation in degrees = 360 minus WXDIFF alpha (TODO:clarify)
- 'splineFile' optional spline file describing detector distortion

WXDIFF FORMAT
-------------

Similar to Fit2D format,
but knowledge about WXDIFF is hard to come by. 
I hope it can be cleanly documented here over time.
Detector plane origin is the bottom left corner of the detector. 

.calib file lines (and notes): 
- imagetype=uncorrected-q           TODO: describe 
- dtype=uint16                      img data type = unsigned 16-bit integers
- horsize=___                       horizontal extent of image, in pixels
- versize=___                       vertical extent of image, in pixels
- region_ulc_x=___                  TODO: describe  
- region_ulc_y=___                  TODO: describe  
- bcenter_x=___                     horizontal coordinate where the beam axis intersects the detector plane 
- bcenter_y=___                     vertical coordinate where the beam axis intersects the detector plane 
- detect_dist=___                   direct distance from the sample to the detector plane intersection, along the beam axis, in pixels
- detect_tilt_alpha=___             rotation of detector tilt axis plane in radians = 360 minus Fit2D tiltPlanRotation
- detect_tilt_delta=___             detector tilt in radians (TODO:clarify)
- wavelenght=___                    the typo 'wavelenght' is built into wxdiff, and it is reported in angstroms
- Qconv_const=___                   TODO: describe 
"""





from collections import OrderedDict

import pyFAI

from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(poni_dict=None)

class ReadPONI(Operation):
    """
    Read in a dict of PyFAI PONI parameters.
    Input path to a .poni file representing a calibrated measurement geometry.
    """
    
    def __init__(self):
        super(ReadPONI,self).__init__(inputs,outputs)
        self.input_doc['file_path'] = 'Path to a .poni file '\
            'describing a calibrated sample-detector geometry'
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters'

    def run(self):
        fpath = self.inputs['file_path']
        g = pyFAI.geometry.Geometry()
        g.read(fpath)
        pdict = g.getPyFAI()
        self.outputs['poni_dict'] = pdict 




from collections import OrderedDict
import os
import time

from ...Operation import Operation

inputs=OrderedDict(
    flow_reactor=None,
    recipe=None,
    delay_time=0,
    delay_volume=0)
outputs=OrderedDict()
 
class SetFlowReactor(Operation):
    """Set the temperature and flow rates for a FlowReactor Plugin."""

    def __init__(self):
        super(SetFlowReactor,self).__init__(inputs,outputs)
        self.input_doc['flow_reactor'] = 'FlowReactor plugin'
        self.input_doc['recipe'] = 'dict describing temperatures and flow rates'
        self.input_doc['delay_time'] = 'minimum time (seconds) to wait after recipe is set'
        self.input_doc['delay_volume'] = 'minimum volume (microlitres) to pump after recipe is set'

    def run(self):
        fr = self.inputs['flow_reactor'] 
        rcp = self.inputs['recipe']
        #self.message_callback(fr.prettyprint_recipe(rcp))
        fr.set_recipe(rcp)
        delvol = self.inputs['delay_volume']
        del_time = self.inputs['delay_time']
        vol_time = 60*delvol/rcp['flowrate']
        if del_time > vol_time:
            self.message_callback('blocking {} seconds'.format(del_time))
            time.sleep(del_time)
            self.message_callback('finished blocking')
        else:
            msg = 'blocking {} seconds to pump {} microlitres'.format(vol_time,delvol)
            self.message_callback(msg)
            time.sleep(vol_time)
            self.message_callback('finished blocking')
        flag,statstr,statdict = fr.check_status()
        self.message_callback(statstr)




from collections import OrderedDict
import time
import os

from ...Operation import Operation

inputs=OrderedDict(
    marccd_client=None,
    exposure_time=10.,
    filename=None)
outputs=OrderedDict()
        
class MarCCDExpose(Operation):
    """Trigger a running MarCCDClient Plugin to take an exposure"""

    def __init__(self):
        super(MarCCDExpose,self).__init__(inputs,outputs)
        self.input_doc['marccd_client'] = 'A MarCCDClient Plugin'
        self.input_doc['exposure_time'] = 'The exposure time in seconds' 
        self.input_doc['filename'] = 'Filename to save the image' 

    def run(self):
        cl = self.inputs['marccd_client'] 
        t_exp = self.inputs['exposure_time'] 
        fn = self.inputs['filename']
        cl.request_exposure(t_exp,fn)




from __future__ import print_function
from collections import OrderedDict

import numpy as np
from xrsdkit import system as xrsdsys

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    source_wavelength=None,
    system=None,
    error_weighted=True,
    logI_weighted=True,
    q_range=[0.,float('inf')]) 

outputs = OrderedDict(
    system=None,
    system_dict=None,
    q_I_opt=None)

class XRSDFit(Operation):
    """Fit scattering equation parameters to measured data.

    Optimizes the parameters of a material system (xrsdkit.system.System) 
    against a measured intensity profile (I(q) vs. q).
    """

    def __init__(self):
        super(XRSDFit, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of wave vectors (1/Angstrom) and intensities (counts)'
        self.input_doc['source_wavelength'] = 'wavelength of light source, in Angstroms'
        self.input_doc['system'] = 'material system description, as a xrsdkit.system.System'
        self.input_doc['error_weighted'] = 'flag for using I(q) error estimates to weight fit objective'
        self.input_doc['logI_weighted'] = 'flag for evaluating the fit objective on log(I)'
        self.input_doc['q_range'] = 'lower and upper q-limits for the fit objective'
        self.output_doc['system'] = 'xrsdkit.system.System with optimized parameters'
        self.output_doc['q_I_opt'] = 'computed intensity for the optimized system'

    def run(self):
        q_I = self.inputs['q_I']
        src_wl = self.inputs['source_wavelength']        
        sys = self.inputs['system']
        err_wtd = self.inputs['error_weighted']
        logI_wtd = self.inputs['logI_weighted']
        q_range = self.inputs['q_range']
        if sys is not None:
            if sys.fit_report:
                err_wtd = sys.fit_report['error_weighted']
                logI_wtd = sys.fit_report['logI_weighted']
                q_range = sys.fit_report['q_range']

        self.message_callback('fitting pattern')
        sys_opt = xrsdsys.fit(sys,q_I[:,0],q_I[:,1],src_wl,None,err_wtd,logI_wtd,q_range)
        self.message_callback('done (objective: {} --> {})'.format(
            sys_opt.fit_report['initial_objective'],sys_opt.fit_report['final_objective']))

        # TODO: update report to new xrsdkit API
        #self.message_callback(xrf.print_report(pops,fit_pops,rpt))
 
        I_opt = sys_opt.compute_intensity(q_I[:,0],src_wl) 
        q_I_opt = np.array([q_I[:,0],I_opt]).T

        self.outputs['system'] = sys_opt 
        self.outputs['system_dict'] = sys_opt.to_dict()
        self.outputs['q_I_opt'] = q_I_opt
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import copy
from functools import partial

from ..pawstools import DictTree 

class Workflow(DictTree):
    """Workflow built from paws Operations, with DictTree interface."""

    def __init__(self,inputs,outputs):
        super(Workflow,self).__init__()

        # TODO: get the proper module_path for the subclasses
        self._module_path = __name__

        self.operations = self._root
        self.ops_enabled = OrderedDict() 
        self.log_file = None
        self.message_callback = self.tagged_print
        self.data_callback = self.set_data
        self.stop_flag = False

        self.declare_inputs(copy.deepcopy(inputs))
        self.declare_outputs(copy.deepcopy(outputs))


    def declare_inputs(self,inputs):
        self.inputs = OrderedDict(copy.deepcopy(inputs))
        self.input_doc = OrderedDict.fromkeys(self.inputs.keys()) 

    def declare_outputs(self,outputs):
        self.outputs = OrderedDict(copy.deepcopy(outputs))
        self.output_doc = OrderedDict.fromkeys(self.outputs.keys()) 

    def add_operations(self,**kwargs):
        for op_name, op in kwargs.items():
            self.add_operation(op_name,op)

    def add_operation(self,op_name,op):
        """Name and add an Operation to the Workflow.
    
        If `op_name` is not unique, 
        the existing Operation is overwritten.
    
        Parameters
        ----------
        op_name : str
            name to give to the new Operation 
        """
        #op.message_callback = self.message_callback
        op.data_callback = partial(self.set_op_item,op_name)
        self.ops_enabled[op_name] = True
        self.set_data(op_name,op)

    def tagged_print(self,msg):
        msg = '[{}] {}'.format(type(self).__name__,msg)  
        print(msg)
        if self.log_file:
            f = open(self.log_file,'a')
            f.write(msg+os.linesep)
            f.close()

    def set_log_file(self,file_path=None):
        if not file_path:
            suffix = 0
            default_file = '{}_{}.log'.format(type(self).__name__,suffix)
            file_path = os.path.join(pawstools.paws_scratch_dir,default_file)
            while os.path.exists(file_path):
                pth,ext = os.path.splitext(file_path)
                file_path = pth+'_{}'.format(suffix)+ext
                suffix += 1
        if file_path == self.log_file:
            return
        open(file_path,'a').close()
        self.message_callback('workflow log file: {}'.format(file_path))
        self.log_file = file_path 

    def set_op_inputs(self,op_name,**kwargs):
        for input_name,val in kwargs.items():
            self.set_op_input(op_name,input_name,val)

    def set_op_input(self,op_name,input_name,val):
        #if not op_name in self.operations:
        #    raise KeyError('operation name {} does not exist'.format(op_name))
        #if not input_name in self.operations[op_name].inputs:
        #    raise KeyError('input name {} does not exist for operation {}'
        #    .format(input_name,op_name))
        itm_key = op_name+'.inputs.'+input_name 
        self.verify_keys([itm_key])
        self.set_data(itm_key,val)

    def set_op_item(self,op_name,item_key,item_data):
        """Subroutine for use with functools.partial for callbacks"""
        full_key = op_name+'.'+item_key
        self.set_data(full_key,item_data)

    def verify_keys(self,keys):
        for k in keys:
            kparts = k.split('.')
            # check only the first three parts of the key:
            # things should be documented to that depth
            if len(kparts) > 3: kparts = kparts[:3]
            k_check = kparts.pop(0)
            for subk in kparts:
                k_check += '.' + subk
            if not k_check in self.keys():
                raise KeyError('Workflow item key {} does not exist'.format(k_check))

    def set_inputs(self,**kwargs):
        for input_name,val in kwargs.items():
            self.inputs[input_name] = val

    #def set_input(self,input_name,val):
    #    """Set a value for Workflow items in self.input_map[`input_name`]"""
    #    r = self.input_map[input_name]
    #    for k in r:
    #        self.set_data(k,val)
    #        # Update the wf_item map so the value is loaded at runtime 
    #        #self.set_wf_item(k,val)

    #def get_output(self,output_name):
    #    out_map = self.outputs[output_name]
    #    if len(out_map) == 1:
    #        return self.get_data(out_map[0])
    #    else:
    #        return [self.get_data(k) for k in out_map]

    #def set_wf_item(self,wf_item_key,val):
    #    self.verify_keys([wf_item_key])
    #    self.op_inputs[wf_item_key] = val

    def run_with(self,**kwargs):
        """Run the Workflow with keyword arguments substituted for the inputs.

        Any keyword arguments that match the Workflow.inputs keys
        are loaded into the Workflow.inputs before calling Workflow.run().
        All relevant results are stored in Workflow.outputs.
        """
        for k in kwargs.keys():
            if not k in self.inputs:
                raise ValueError('Input {} is not valid for Workflow {}'.format(k,type(self).__name__))
        self.inputs.update(kwargs)
        return self.run() 

    def run(self):
        """Run the Workflow."""
        pass

    def stop(self):
        """Stop the Workflow and all of its Operations."""
        with self.stop_lock:
            self.stop_flag = True
        # TODO: revisit this after implementing workflow threading
        #for op_name,op in self.operations.items():
        #    op.stop()

    def build_clone(self):
        """Produce a clone of this Workflow."""
        new_wf = self.clone() 
        new_wf.inputs = copy.deepcopy(self.inputs)
        new_wf.outputs = copy.deepcopy(self.outputs)
        # TODO: consider how to handle callbacks; mind threading 
        new_wf.message_callback = self.message_callback
        new_wf.data_callback = self.data_callback
        return new_wf

    def setup_dict(self):
        """Return a dict that describes the Workflow setup.""" 
        wf_dict = OrderedDict()
        wf_dict['WF_MODULE'] = self._module_path 
        wf_dict['INPUTS'] = copy.deepcopy(self.inputs)
    #    wf_dict['OUTPUTS'] = copy.deepcopy(self.outputs)
        return wf_dict

    @classmethod
    def clone(cls):
        return cls()

    def enable_ops(self,*args):
        for op_name in args:
            self.enable_op(op_name)

    def disable_ops(self,*args):
        for op_name in args:
            self.disable_op(op_name)

    def enable_op(self,op_name):
        self.set_op_enabled(op_name,True)

    def disable_op(self,op_name):
        self.set_op_enabled(op_name,False)

    def set_op_enabled(self,op_name,flag=True):
        self.ops_enabled[op_name] = flag 




from __future__ import print_function
from collections import OrderedDict
import re
#from multiprocessing import Process,Pool

import yaml

from ..operations.OpManager import OpManager
from ..plugins.PluginManager import PluginManager
from .. import pawstools
from . import import_workflow

class WfManager(object):
    """Manager for paws Workflows."""

    def __init__(self,op_manager=None,plugin_manager=None):
        """Initialize a workflow manager.

        Parameters
        ----------
        op_manager : OpManager (optional)
            an operations manager (see paws.operations.OpManager)-
            if not provided, a default OpManager will be created.
        plugin_manager : PluginManager (optional)
            a plugins manager (see paws.plugins.PluginManager)-
            if not provided, a default PluginManager will be created.
        """
        super(WfManager,self).__init__()
        if not op_manager:
            op_manager = OpManager()
        if not plugin_manager:
            plugin_manager = PluginManager()
        self.op_manager = op_manager 
        self.plugin_manager = plugin_manager 
        self.workflows = OrderedDict() 
        # dict of workflow clones for executing across threads:
        self.wf_clones = OrderedDict()
        # dict of bools to keep track of who is at work:
        self.wf_running = OrderedDict() 
        self.message_callback = self.tagged_print

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def add_workflow(self,wf_name,wf_module):
        """Name and add a workflow.
    
        If `wf_name` is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
    
        Parameters
        ----------
        wf_name : str
            name to give to the new Workflow
    
        Returns
        -------
        wf : Workflow
            a reference to the new Workflow
        """
        wf = import_workflow(wf_module)() 
        if not wf.is_key_valid(wf_name): 
            raise KeyError(wf.key_error_message(wf_name))
    #    #wf.message_callback = self.message_callback
        self.workflows[wf_name] = wf
        self.wf_running[wf_name] = False
        return wf

    #def add_operation(self,wf_name,op_name,op_module_path):
    #    """Name and add an Operation to a Workflow.
    #
    #    Parameters
    #    ----------
    #    wf_name : str
    #        name of the Workflow to add the Operation to
    #    op_name : str
    #        name to give to the new Operation
    #    op_module_path : str
    #        path to locate the Operation module (e.g. CAT.SUBCAT.OpName)
    #    """
    #    self.workflows[wf_name].add_operation(op_name,self.get_operation(op_module_path))

    #def get_operation(self,op_module_path):
    #    """Get the Operation at `op_module_path` from self.op_manager"""
    #    return self.op_manager.get_operation(op_module_path)

    def run_workflow(self,wf_name,inputs={}):
        """Execute the workflow indicated by `wf_name`.

        The workflow is cloned, and the cloned workflow is executed.
        As it runs, depending on the implementation of Workflow.run(),
        the clone may update the original Workflow's data.

        Parameters
        ----------
        wf_name : str
            Name of the workflow to be executed,
            must be one of the keys in self.workflows.

        Returns
        -------
        wf_clone.outputs : dict
            Dict containing the cloned workflow's output data.
        """
        # TODO: support running these off the main thread:
        # implement running_lock for running flag, 
        # data_lock (use in data_callback) for updating inputs/outputs
        #wf = self.workflows[wf_name]
        #self.message_callback('preparing workflow {} for execution'.format(wf_name))
        #stk,diag = wf.execution_stack()
        #wf_clone = self.prepare_wf(wf_name,inputs)
        #wf_clone.data_callback = wf.data_callback
        self.wf_clones[wf_name] = self.workflows[wf_name].build_clone()  
        self.wf_running[wf_name] = True
        self.wf_clones[wf_name].run()
        self.message_callback('execution finished')
        return wf_clone.outputs

    def stop_workflow(self,wf_name):
        """Stop the workflow indicated by `wf_name`"""
        self.message_callback('stopping workflow {}'.format(wf_name))
        if wf_name in self.wf_clones.keys():
            wf = self.wf_clones.pop(wf_name)
            wf.stop()
        self.wf_running[wf_name] = False

    def load_workflow(self,wf_name,wf_dict):
        """Load a workflow from a dict that specifies its parameters.
        
        If `wf_name` is not unique, self.workflows[wf_name] is overwritten.
        
        Parameters
        ----------
        wf_name : str
            name to be given to the new workflow
        wf_dict : dict
            dict specifying workflow setup
        """
        self.add_workflow(wf_name,wf_dict['WF_MODULE'])
        for inpname,val in wf_dict['INPUTS'].items():
            self.workflows[wf_name].inputs[inpname] = val
        #for op_tag, op_module in wf_dict['OPERATIONS'].items():
        #    self.load_operation(wf_name,op_tag,op_module)
        #for outname,out_map in wf_dict['OUTPUTS'].items():
        #    self.workflows[wf_name].connect_output(outname,out_map)
        #for input_key,val in wf_dict['OP_INPUTS'].items():
        #    self.workflows[wf_name].set_wf_item(input_key,val)
        #for out_key, in_map in wf_dict['CONNECTIONS'].items():
        #    self.workflows[wf_name].connect(out_key,in_map)
        #for op_name, dep_ops in wf_dict['DEPENDENCIES'].items():
        #    self.workflows[wf_name].set_dependency(op_name,dep_ops)
        #for op_name, flag in wf_dict['ENABLED_FLAGS'].items():
        #    self.workflows[wf_name].set_op_enabled(op_name,flag)
        #for plugin_item_key, input_map in wf_dict['PLUGIN_CONNECTIONS'].items():
        #    self.workflows[wf_name].connect_plugin(plugin_item_key,input_map)
        #for wwffnm, input_map in wf_dict['WORKFLOW_CONNECTIONS'].items():
        #    self.workflows[wf_name].connect_workflow(wwffnm,input_map)

    def load_wfm(self,wfm_filename):
        """Set up the WfManager and its OpManager and PluginManager, from a .wfm file.
    
        Parameters
        ----------
        wfm_filename : str
            path to a .wfm file to be loaded
        """
        f = open(wfm_filename,'r')
        d = yaml.load(f)
        f.close()
        if 'PAWS_VERSION' in d.keys():
            wfm_version = d['PAWS_VERSION']
        else:
            wfm_version = '0.0.0'
        wfm_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',wfm_version)
        wfm_vparts = list(map(int,wfm_vparts.groups()))
        current_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',pawstools.__version__)  
        current_vparts = list(map(int,current_vparts.groups()))
        if wfm_vparts[0] < current_vparts[0] or wfm_vparts[1] < current_vparts[1]:
            warnings.warn('WARNING: paws (version {}) does not match file version ({})'\
            .format(pawstools.__version__,wfm_version))  
        if 'OP_ENABLED_FLAGS' in d.keys():
            for op_module,flag in d['OP_ENABLED_FLAGS'].items():
                self.op_manager.enable_op(op_module)
        if 'WORKFLOWS' in d.keys():
            wf_dict = d['WORKFLOWS']
            for wf_name,wf_setup_dict in wf_dict.items():
                self.load_workflow(wf_name,wf_setup_dict)
        if 'PLUGINS' in d.keys():
            self.plugin_manager.load_plugins(d['PLUGINS'])

    def load_operations(self,*args):
        return [self.load_operation(opmd) for opmd in args]
    
    def load_operation(self,op_module):
        """Load an Operation from a dict that specifies its parameters.

        If `op_name` is not unique, the Operation is overwritten.

        Parameters
        ----------
        op_module : str 
            key for locating Operation module 
        """
        if not self.op_manager.is_op_enabled(op_module):
            self.op_manager.enable_op(op_module)
        op = self.op_manager.get_data(op_module)()
        return op

    def setup_dict(self):
        d = {} 
        d['PAWS_VERSION'] = pawstools.__version__
        d['OP_ENABLED_FLAGS'] = {k:True for k in self.op_manager.list_operations() if self.op_manager.is_op_enabled(k)}
        wfman_dict = OrderedDict.fromkeys(self.workflows.keys())
        for wfname,wf in self.workflows.items():
            wfman_dict[wfname] = wf.setup_dict()
        d['WORKFLOWS'] = wfman_dict
        d['PLUGINS'] = self.plugin_manager.setup_dict()
        return d
    
    #    wf_dict['OPERATIONS'] = OrderedDict.fromkeys(wf.operations)
    #    for op_name,op in wf.operations.items():
    #        wf_dict['OPERATIONS'][op_name] = op.__module__[op.__module__.find('operations.')+11:] 
    #    wf_dict['OP_INPUTS'] = wf.op_inputs
    #    wf_dict['CONNECTIONS'] = wf.op_connections
    #    wf_dict['DEPENDENCIES'] = wf.op_dependencies
    #    wf_dict['ENABLED_FLAGS'] = wf.op_enabled_flags()
    #    wf_dict['PLUGIN_CONNECTIONS'] = wf.plugin_connections
    #    wf_dict['WORKFLOW_CONNECTIONS'] = wf.workflow_connections

    #def save_to_wfm(self,wfm_filename):
    #    """Save workflows, plugins, and active operations to a .wfm file.
    #
    #    The .wfm file is really just a YAML file. 
    #
    #    Parameters
    #    ----------
    #    wfm_filename : str
    #        full path of the .wfm file to be saved-
    #        extension is automatically appended if not provided, 
    #        and an existing file will be overwritten.
    #    """
    #    if not os.path.splitext(wfm_filename)[1] == '.wfm':
    #        wfm_filename = wfm_filename + '.wfm'
    #    #print('saving workflow manager setup to {}'.format(wfm_filename))
    #    pawstools.save_file(wfm_filename,self.setup_dict())

    #def save_to_wfl(self,wf_name,wfl_filename):
    #    if not os.path.splitext(wfl_filename)[1] == '.wfl':
    #        wfl_filename = wfl_filename + '.wfl'
    #    #print('saving {} to {}'.format(wf_name,wfl_filename))
    #    pawstools.save_file(wfl_filename,self.wf_setup_dict(wf_name))

    ## TODO: unify workflow load/save architecture
    #
    #def load_packaged_wfm(self,wf_key):
    #    # the following import saves a .wfm configuration file 
    #    #importlib.import_module('.'+workflow_key,pawstools.wf_module)
    #    import_workflow_module(wf_key)
    #    wfm_path = pawstools.sourcedir
    #    wfm_path = os.path.join(wfm_path,'workflows')
    #    p = wf_key.split('.')
    #    for mp in p:
    #        wfm_path = os.path.join(wfm_path,mp)
    #    wfm_filename = wfm_path+'.wfm'
    #    self.load_wfm(wfm_filename)

    #def load_packaged_workflow(self,wf_name,wf_key):
    #    # the following import saves a .wfl configuration file 
    #    #importlib.import_module('.'+wf_key,pawstools.wf_module)
    #    import_workflow_module(wf_key)
    #    wf_path = pawstools.sourcedir
    #    wf_path = os.path.join(wf_path,'workflows')
    #    p = wf_key.split('.')
    #    for mp in p:
    #        wf_path = os.path.join(wf_path,mp)
    #    wf_filename = wf_path+'.wfl'
    #    f = open(wf_filename,'r')
    #    d = yaml.load(f)
    #    f.close()
    #    self.load_workflow(wf_name,d)

    #def prepare_wf(self,wf_name,inputs):
    #    wf_clone = self.workflows[wf_name].build_clone()
    #    #for pgn_itm_key,input_map in wf_clone.plugin_connections.items():
    #    #    pgn_itm = self.plugin_manager.get_data(pgn_itm_key)
    #    #    for input_key in input_map:
    #    #        wf_clone.set_data(input_key,pgn_itm)
    #    #for wf_name,input_map in wf_clone.workflow_connections.items():
    #    #    stk,diag = self.workflows[wf_name].execution_stack()
    #    #    new_wf = self.prepare_wf(wf_name,stk)
    #    #    for input_key in input_map:
    #    #        wf_clone.set_data(input_key,new_wf)
    #        # TODO: think about appropriate way for these workflows to callback,
    #        # keep in mind they may be batch-executed, maybe in parallel 
    #        #new_wf.message_callback = self.workflows[wf_name].message_callback
    #        #new_wf.data_callback = self.workflows[wf_name].data_callback
    #    return wf_clone





from .. import TimeSeriesXRSDFitGUI as TimeSeriesXRSDFitGUI_new
from ....IO.BL15.LEGACY.ReadTimeSeries import ReadTimeSeries 

class TimeSeriesXRSDFitGUI(TimeSeriesXRSDFitGUI_new.TimeSeriesXRSDFitGUI):

    def __init__(self):
        super(TimeSeriesXRSDFitGUI,self).__init__()
        self.add_operation('read_timeseries',ReadTimeSeries())




import numpy as np
from collections import OrderedDict

import pyFAI

from ...Operation import Operation

inputs = OrderedDict(
    image_data=None,
    integrator=None,
    npt_rad=1000,
    npt_azim=1000,
    polz_factor=0.,
    unit='q_A^-1',
    integrate_args={})
outputs = OrderedDict(
    q=None,
    chi=None,
    I_at_q_chi=None)

class ApplyIntegrator2d(Operation):
    """Integrate an image using an existing PyFAI.AzimuthalIntegrator.
    
    Input image data (ndarray), PyFAI.AzimuthalIntegrator, 
    and a dict of other keyword parameters to be passed to
    PyFAI.AzimuthalIntegrator.integrate1d().
    
    Refer to the PyFAI documentation at 
    http://pyfai.readthedocs.io/en/latest/ 
    for supported keyword arguments 
    as well as parameter definitions and defaults.
    """

    def __init__(self):
        super(ApplyIntegrator2d,self).__init__(inputs,outputs)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['integrator'] = 'A PyFAI.AzimuthalIntegrator object'
        self.input_doc['npt_rad'] = 'number of q-points to integrate'
        self.input_doc['npt_azim'] = 'number of chi-points to integrate'
        self.input_doc['polz_factor'] = 'polarization factor, '\
            'in case polarization correction is needed'
        self.input_doc['unit'] = 'choice of unit. See PyFAI documentation for options.' 
        self.input_doc['integrate_args'] = 'dict of keyword args '\
            'to pass to pyFAI.AzimuthalIntegrator.integrate2d(). '\
            'Where relevant, the values in this dict will be replaced '\
            'by inputs to the operation, e.g. for npt_rad and npt_azim.' 
        self.output_doc['q'] = 'Scattering vector magnitude q array in 1/Angstrom.'
        self.output_doc['chi'] = 'Azimuthal angle array.'
        self.output_doc['I_at_q_chi'] = '2d array of integrated intensity at q,chi.'

    def run(self):
        img = self.inputs['image_data']
        intgtr = self.inputs['integrator']
        npt_rad = self.inputs['npt_rad']
        npt_azim = self.inputs['npt_azim']
        kw = self.inputs['integrate_args']
        if self.inputs['polz_factor']:
            kw['polarization_factor'] = self.inputs['polz_factor']
        if self.inputs['unit']:
            kw['unit'] = self.inputs['unit']

        I_at_q_chi,q,chi = intgtr.integrate2d(img,npt_rad,npt_azim,**kw)
        
        self.outputs['q'] = q
        self.outputs['chi'] = chi
        self.outputs['I_at_q_chi'] = I_at_q_chi




from collections import OrderedDict
import time
import os

from ...Operation import Operation

inputs=OrderedDict(
    spec_infoclient=None,
    n_loops=1,
    exposure_time=10.,
    delay_time=0.,
    dir_path=None,
    filename=None,
    block=False)
outputs=OrderedDict()
        
class LoopScan(Operation):
    """Run a loop scan through a SpecInfoClient"""

    def __init__(self):
        super(LoopScan,self).__init__(inputs,outputs)
        self.input_doc['spec_infoclient'] = 'A SpecInfoClient'
        self.input_doc['n_loops'] = 'Number of exposures to run' 
        self.input_doc['exposure_time'] = 'The exposure time in seconds' 
        self.input_doc['delay_time'] = 'Time to wait after sending loop scan commands' 
        self.input_doc['dir_path'] = 'Path to where images should be saved (on SPEC PC)' 
        self.input_doc['filename'] = 'Filename root for the images' 
        self.input_doc['block'] = 'If True, block until loop scan completes' 

    def run(self):
        cl = self.inputs['spec_infoclient'] 
        n_loops = self.inputs['n_loops'] 
        t_exp = self.inputs['exposure_time'] 
        t_delay = self.inputs['delay_time'] 
        dp = self.inputs['dir_path']
        fn = self.inputs['filename']
        blk = self.inputs['block']
        #mroot = os.path.join(dp,fn)
        mroot = dp+'/'+fn   # assume the SPEC PC is Unix
        cl.mar_enable()
        cl.run_loopscan(mroot,n_loops,t_exp,blk)
        #if t_delay > 0.:
        #    self.message_callback('delaying {} seconds...'.format(t_delay))
        #    time.sleep(t_delay)
        cl.mar_disable()






from collections import OrderedDict
import glob

from ...Operation import Operation

inputs=OrderedDict(
    dir_path=None,
    regex='*',
    new_files_only=False)
outputs=OrderedDict(file_iterator=None)

class FileIterator(Operation):
    """Iterate over files matching a regex in a directory"""

    def __init__(self):
        super(FileIterator,self).__init__(inputs,outputs)
        self.input_doc['dir_path'] = 'path to directory containing files'
        self.input_doc['regex'] = 'regular expression to select files'
        self.input_doc['new_files_only'] = 'regular expression to select files'

    def run(self):
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        process_existing_files = not self.inputs['new_files_only']
        it = FileSystemIterator(dirpath,rx,process_existing_files) 
        self.outputs['file_iterator'] = it

class FileSystemIterator(Iterator):

    def __init__(self,dirpath,regex,include_existing_files=True):
        self.dirpath = dirpath
        self.rx = regex
        self.paths_done = []
        if not include_existing_files:
            self.paths_done = glob.glob(self.dirpath+'/'+self.rx)
        super(FileSystemIterator,self).__init__()

    def next(self):
        batch_list = glob.glob(self.dirpath+'/'+self.rx)
        for path in batch_list:
            if not path in self.paths_done:
                self.paths_done.append(path)
                return path
        return None

    def __next__(self):
        return self.next()





from __future__ import print_function
from collections import OrderedDict
import copy
import os
from threading import Condition
import serial

import numpy as np
from scipy.optimize import minimize as scipimin

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    serial_device=None,
    timer=None,
    flowrate_table=None,
    state={})

class MitosPPumpController(PawsPlugin):
    """PAWS Plugin for controlling a Mitos P-pump.

    Uses a serial port (or virtual serial port)
    to communicate to a USB-attached Mitos P-pump.
    RS232 protocol settings: 
    57600 baud, 8 data bits, 1 stop bit, no parity, no handshaking.
    The pump controller acts as a proxy for controlling a clone of itself.
    The clone operates in its own thread, exchanging information 
    with the original pump controller in a threadsafe way,
    at each tick of the timer.

    When the pump is in CONTROL mode,
    a command must be sent every 30 seconds,
    or the pump will automatically exit CONTROL mode.
    """
    def __init__(self):
        super(MitosPPumpController,self).__init__(content)
        self.content_doc['serial_device'] = 'string serial device indicator '\
            '(e.g. "COM2") or filesystem path (e.g. "/dev/ttyUSB0")'
        self.content_doc['timer'] = 'Timer PawsPlugin, used to initiate '\
            'thread-safe data exchanges between the pump controller '\
            'and its operational clone.'
        self.content_doc['flowrate_table'] = 'table mapping flowrates '\
            'to Mitos instrument setpoints (with flow meter calibrated for water). '\
            'This table should be formatted as a list '\
            'containing a list of flowrates and a list of setpoints: '\
            '[[flowrates],[setpoints]], with each specified in microlitres/minute, '\
            'with linear interplation as needed (and no extrapolation).'
        self.ErrorsDict = { 0:'None', \
                            1:'Supply than maximum', \
                            2:'Tare time out', \
                            3:'Tare supply still connected ', \
                            4:'Control start time out', \
                            5:'Pressure target too low', \
                            6:'Pressure target too high', \
                            7:'Leak test supply pressure too low', \
                            8:'Leak test time out', \
                            100:'Broken'}
        self.StateDict  = { 0:'IDLE', \
                            1:'CONTROLLING', \
                            2:'TARE', \
                            3:'ERROR', \
                            4:'LEAK TEST'}
        self.serial_lock = Condition()
        self.ser = None
        self.state_lock = Condition()
        self.content['state'] = OrderedDict(
            error_code = None,
            state_code = None, 
            chamber_pressure = None, 
            supply_pressure = None, 
            target_pressure = None, 
            flow_rate = None, 
            target_flow_rate = None)
        self.thread_blocking = True
        self.flow_conversion_power = 1.

    def calibrate(self):
        frtab = self.content['flowrate_table']
        if frtab is not None:
            self.flow_setpts = frtab[:,0]
            self.flow_meas = frtab[:,1]
            fit_obj = lambda a: np.sum([(fset**a-fmeas)**2 for fset,fmeas in zip(self.flow_setpts,self.flow_meas)])
            res = scipimin( fit_obj,1. )
            self.flow_conversion_power = res.x[0]

    def get_setpt(self,rate):
        if rate == 0.: return 0.
        abs_rate = abs(rate)**(1./self.flow_conversion_power)
        if rate < 0:
            return -1*abs_rate
        else:
            return abs_rate

    def get_true_flowrate(self,setpt):
        if setpt == 0.: return 0.
        abs_setpt = abs(setpt)**self.flow_conversion_power 
        if setpt < 0:
            return -1*abs_setpt
        else:
            return abs_setpt

    def start(self):
        self.calibrate()
        super(MitosPPumpController,self).start()
        with self.thread_clone.running_lock:
            self.thread_clone.running_lock.wait()
       
    def run(self):
        # this method is run by self.thread_clone 
        self.ser = serial.Serial(
            self.content['serial_device'], 
            57600, timeout=1, 
            parity = serial.PARITY_NONE, 
            bytesize = serial.EIGHTBITS, 
            xonxoff = 0, rtscts = 0)
        tmr = self.content['timer']
        keep_going = True
        # check for control...
        self.update_status()
        if not self.content['state']['state_code'] == 1:
            # attempt to enter remote control mode 
            resp = self.run_cmd('A1')
            if not resp == "#A0":
                msg = "Pump failed to enter REMOTE control mode"
                self.message_callback(msg)
                with self.proxy.history_lock:
                    with tmr.dt_lock:
                        t_now = float(tmr.dt_utc())
                    self.proxy.add_to_history(t_now,'ERROR: '+msg)
                keep_going = False
                with self.proxy.running_lock:
                    self.proxy.stop()
                #raise RuntimeError(msg)

        # clear the pump...
        self.run_cmd('C')

        self.run_notify()
        while keep_going:
            with tmr.dt_lock:
                tmr.dt_lock.wait()
            #while self.commands.qsize() > 0: 
            #    with self.command_lock:
            #        cmd = self.commands.get()
            #        self.commands.task_done()
            #    self.run_cmd(cmd)
            self.update_status()
            with tmr.running_lock:
                if not tmr.running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)

        # relinquish control
        self.run_cmd('A0') 
        if self.verbose: self.message_callback('FINISHED')
        with self.proxy.history_lock:
            self.proxy.dump_history()
        self.ser.close()

    def update_status(self):
        with self.state_lock:
            self.read_status()
        with self.proxy.state_lock:
            self.proxy.content['state'] = copy.deepcopy(self.content['state'])
        if self.verbose: self.message_callback(self.print_status())

    def run_cmd(self,cmd):
        with self.content['timer'].dt_lock:
            t_now = float(self.content['timer'].dt_utc())
        with self.serial_lock:
            self.send_line(cmd)
            resp = self.receive_line()
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd+' '+resp)
        return resp

    def send_line(self,line):
        self.ser.write("{}\r\n".format(line).encode('utf-8'))  

    def receive_line(self):
        return self.ser.readline().strip().decode()

    def read_status(self):
        """ Read pump status.

        The reply from the pump is formatted as 
        #sErr,Sp,Src,Pc,Ps,Pt,Qc,Qt,Ft
        These are numeric codes defining the response state.  

        Err - Error code. A value of 0 means CMD_ACCEPTED - OK.

        Sp  - State of the pump
            0.   IDLE
            1.   CONTROLLING
            2.   TARE
            3.   ERROR
            4.   LEAK TEST
        The status returned by the pump should reflect
        the latest command sent from the PC to the pump.

        Src - Mode of control: 0 is manual, 1 is remote controlled.

        Pc  - Chamber pressure in mbar.

        Ps  - Supply pressure in mbar.

        Pt  - Target pressure in mbar. 

        Qc  - Current flow rate in pl/s. 
            Requires a connected Mitos Sensor Display flow meter.

        Qt  - Target flow rate in pl/s. 
            Requires a connected Mitos Sensor Display flow meter.

        Ft  - Flow sensor type plus flow control mode. 
            Requires a connected Mitos Sensor Display,
            or sensor interface flow meter.
            This consists of an upper nibble, 
            middle nibble and lower nibble.
        """
        resp = self.run_cmd('s')
        s = resp.split(',')
        while '' in s or len(s) < 8:
            resp = self.run_cmd('s')
            s = resp.split(',')
        with self.state_lock:
            st = self.content['state']
            st['error_code'] = int(s[0][2:])
            st['state_code'] = int(s[1])
            st['chamber_pressure'] = float(s[3])
            st['supply_pressure'] = float(s[4])
            st['target_pressure'] = float(s[5])
            st['flow_rate'] = float(s[6])
            st['target_flow_rate'] = float(s[7])

    def print_status(self):
        with self.state_lock:
            st = self.content['state']
            msg = os.linesep+'----------- PUMP STATUS -----------'+os.linesep
            msg += 'Error code: {} ({})'.format(st['error_code'],
                self.ErrorsDict[st['error_code']])+os.linesep
            msg += 'State code: {} ({})'.format(st['state_code'],
                self.StateDict[st['state_code']])+os.linesep
            msg += 'Chamber pressure: \t{}'.format(st['chamber_pressure'])+os.linesep 
            msg += 'Supply pressure: \t{}'.format(st['supply_pressure'])+os.linesep 
            msg += 'Target pressure: \t{}'.format(st['target_pressure'])+os.linesep 
            msg += 'Current flow rate: \t{}'.format(st['flow_rate'])+os.linesep 
            msg += 'Target flow rate: \t{}'.format(st['target_flow_rate'])+os.linesep 
            msg += '-----------------------------------'
        return msg

    def set_flowrate(self,rate):
        """Set the pump flow rate to `rate` (microlitres per minute).

        Parameters
        ----------
        rate : float 
            The flow rate set point, in microlitres per minute.
        """
        # TODO: think of an elegant way to handle setpts below the resolution of the pump controls
        if rate < 0.1:
            rate = 0.
        if self.verbose: self.message_callback('requested flowrate: {}'.format(rate))
        setpt = self.get_setpt(rate) 
        if self.verbose: self.message_callback('calibrated setpoint for requested flowrate: {}'.format(setpt))
        pl_s_rate = int(round(float(setpt)*1.E6/60.))
        if self.verbose: self.message_callback('calibrated setpoint in pL/sec: {}'.format(pl_s_rate))
        self.thread_clone.run_cmd('F{}'.format(pl_s_rate))
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put("F{}".format(pl_s_rate)) 

    def set_pressure(self,pressure):
        """Set the pump pressure to the provided value.
    
        Parameters
        ----------
        pressure : integer
            The pressure set point, in mbar.
            A value of 1000 sends the 'P1000' command,
            which sets target pressure to 1000 mbar. 
        """
        self.thread_clone.run_cmd('P{}'.format(pressure))
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put("P{}".format(pressure)) 

    def tare(self):
        """Tare the P-pump."""
        if self.verbose: self.message_callback('taring pump')
        self.thread_clone.run_cmd('R0')
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put("R0")

    def set_idle(self):
        """Set the P-pump to idle."""
        if self.verbose: self.message_callback('setting pump to idle')
        self.thread_clone.run_cmd('P0')
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put("P0")

    def dispense_volume(self, vol):
        """Control the pump to dispense a specified volume.

        Parameters
        ----------
        vol : float
            The volume to dispense 
        """
        pass




from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs = OrderedDict(
    x=None,
    y=None,
    dy=None,
    order=None,
    base=None)
outputs = OrderedDict(smoothdata=None)

class SavitzkyGolay(Operation):
    """Applies a Savitzky-Golay polynomial smoothing filter to a 1d array."""

    def __init__(self):
        super(SavitzkyGolay, self).__init__(inputs, outputs)
        self.input_doc['x'] = '1d array- independent variable'
        self.input_doc['y'] = '1d array- dependent variable, same shape as x'
        self.input_doc['dy'] = '1d array, error estimate in y, same shape as y (default None)'
        self.input_doc['order'] = 'integer order of polynomial approximation (zero to five)'
        self.input_doc['base'] = '-1, 0, or positive integer; see class docs'
        self.output_doc['smoothdata'] = 'smoothed 1d array for y'

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        o = self.inputs['order']
        b = self.inputs['base']
        err = self.inputs['dy']
        nx = x.size
        if err is None:
            err = np.ones(y.shape,dtype=float)        
        # TODO: document what is being done here.
        # "Minimal" point base case. 
        if b == -1:
            o_odd_flag = ((o/2)*2 != o)
            npts = o+1+int(o_odd_flag)
        # "Balanced" point base case.
        elif b == 0:
            npts = 2*o+1
        # "Additional" point base case.
        elif b > 0:
            npts = 2*(o+b)+1
        # TODO: build y_out in a list comprehension.
        y_out = np.zeros(nx, dtype=float)
        for i in range(nx):
            # choose start and end indices
            npts_half = int(npts)/2 + 1
            start = max([i-npts_half+1,0])
            end = min([i+npts_half-1,nx-1])
            if b == -1:
                if (i-npts_half+1 < 0):
                    start = 0
                    end = npts 
                elif (i+npts_half-1 > nx-1):
                    start = size-1-m
                    end = size-1
            # Formulate the equation to be solved for polynomial coefficients
            xvals = x[start:end]
            yvals = y[start:end]
            errvals = err[start:end]
            nxvals = xvals.size
            # TODO: document what is happening here
            idxs = np.arange(o+1)
            # reshape vectors for matrix mult
            y_vec = np.reshape(yvals,(nxvals,1,1))
            x_vec = np.reshape(xvals,(nxvals,1,1))
            err_vec = np.reshape(errvals,(nxvals,1,1))
            idx_column = np.reshape(idxs,(nxvals,1)) 
            idx_row = np.reshape(idxs,(1,nxvals)) 
            idx_block = idx_row + idx_column
            v = (y_vec * x_vec ** idx_column * err_vec).sum(axis=0)
            m = (x_vec ** idx_block * err_vec).sum(axis=0)
            # CALL LINALG.SOLVE...
            coefs = (np.linalg.solve(m, v)).flatten()
            y_out[i] = np.sum(x[i]**np.arange(o+1)*coefs)
        self.outputs['smoothdata'] = y_out




from collections import OrderedDict

from ..Operation import Operation

inputs = OrderedDict(n_primes=10)
outputs = OrderedDict(primes_list=[])

class ListPrimes(Operation):
    """Makes a list of prime numbers in increasing order"""

    def __init__(self):
        super(ListPrimes,self).__init__(inputs,outputs) 
        self.input_doc['n_primes'] = 'number of primes to output'
        self.output_doc['primes_list'] = 'list of prime numbers'
    
    def run(self):
        np = self.inputs['n_primes']
        pl = self.outputs['primes_list'] 
        npfound = 0
        x = 2
        while npfound < np:
            divs = [p for p in pl if x%p == 0]
            if not divs:
                npfound += 1 
                pl.append(x)
            x += 1
        return self.outputs




from collections import OrderedDict

import numpy as np

from ..Operation import Operation

inputs = OrderedDict(x_y_arrays=[])
outputs = OrderedDict(x_ymean=None)

class ArrayYMean(Operation):
    """
    Average the second column of one or more n-by-2 arrays
    """

    def __init__(self):
        super(ArrayYMean, self).__init__(inputs,outputs)
        self.input_doc['x_y_arrays'] = 'list of n-by-2 arrays'
        self.output_doc['x_ymean'] = 'n-by-2 array of x and mean(y)'

    def run(self):
        x_y_arrays = self.inputs['x_y_arrays']
        x_ymean = None
        if len(x_y_arrays) > 0:
            x_ymean = np.zeros(x_y_arrays[0].shape)
            x_ymean[:,0] = x_y_arrays[0][:,0]
            x_ymean[:,1] = np.mean([xy[:,1] for xy in x_y_arrays],axis=0)
        self.outputs['x_ymean'] = x_ymean 
        return self.outputs




from __future__ import print_function
import os
import pkgutil
import importlib

def load_workflows(path_,pkg,cat_root=''):
    cat_list = []
    cat_wf_list = []
    wf_modules = {} 
    mods = pkgutil.iter_modules(path_)
    mods = [mod for mod in mods if mod[1] not in ['__init__','Workflow','WfManager','wftools']]
    for modloader, modname, ispkg in mods:
        if ispkg:
            pkg_path = [os.path.join(path_[0],modname)]
            subcat_root = modname
            if cat_root:
                subcat_root = cat_root+'.'+modname
            pkg_cats, pkg_wfs, pkg_wf_modules = load_workflows(pkg_path,pkg+'.'+modname,subcat_root)
            pkg_wfs = [wf for wf in pkg_wfs if not wf in cat_wf_list]
            pkg_cats = [cat for cat in pkg_cats if not cat in cat_list]
            cat_wf_list = cat_wf_list + pkg_wfs
            cat_list = cat_list + pkg_cats
            wf_modules.update(pkg_wf_modules)
        else:
            # assume the module defines workflows and generates a .wfm or .wfl
            if not cat_root in cat_list:
                cat_list.append(cat_root)
            cat_wf_list.append( (cat_root,modname) )
            wf_modules[cat_root+'.'+modname] = os.path.join(path_[0],modname) 
    # don't import these modules: it will cause an import loop
    #for wf_uri,wf_path in wf_modules.items():
    #    wf_mod = importlib.import_module('.'+wf_uri,__name__)
    return cat_list, cat_wf_list, wf_modules

cat_list, cat_wf_list, wf_modules = load_workflows(__path__,__name__)

def import_workflow(wf_module_key):
    mod = importlib.import_module('.'+wf_module_key,__name__)
    return mod.__dict__[wf_module_key.split('.')[-1]]

#print(cat_list)
#print(cat_wfl_list)
#print(wfl_modules)




from collections import OrderedDict
import os

import numpy as np

from ..Operation import Operation

inputs=OrderedDict(
    file_path=None,
    dtype='float',
    delimiter=' ')

outputs=OrderedDict(
    data=None,
    dir_path=None,
    filename=None) 

class NumpyLoad(Operation):
    """Wrapper for numpy.loadtxt."""

    def __init__(self):
        super(NumpyLoad, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to data file'
        self.input_doc['dtype'] = 'datatype (given as a string)'
        self.input_doc['delimiter'] = 'delimiter (str) placed between values'
        self.output_doc['data'] = 'result of loading file contents'
        self.output_doc['dir_path'] = 'directory of the input file' 
        self.output_doc['filename'] = 'name of the input file, no path, no extension' 
        self.output_doc['extension'] = 'extension of the input file path' 

    def run(self):
        p = self.inputs['file_path']
        self.message_callback('reading {}'.format(p))
        dlm = self.inputs['delimiter']
        dtp = eval(self.inputs['dtype']) 
        self.outputs['dir_path'], fn = os.path.split(p)
        self.outputs['filename'], self.outputs['extension'] = os.path.splitext(fn) 
        self.outputs['data'] = np.loadtxt(p, dtype=dtp, delimiter=dlm)
        return self.outputs




from collections import OrderedDict

from xrsdkit.visualization.gui import run_fit_gui
import numpy as np

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    system=None,
    error_weighted=True,
    logI_weighted=True,
    q_range=[0.,float('inf')]
    )

outputs = OrderedDict(
    system=None,
    system_dict=None,
    q_I_opt = None
    ) 

class XRSDFitGUI(Operation):
    """Interactively fit a XRSD spectrum."""

    def __init__(self):
        super(XRSDFitGUI, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q(1/Angstrom) versus I(arb).'
        self.input_doc['system'] = 'material system description, as a xrsdkit.system.System'
        self.input_doc['error_weighted'] = 'flag for using I(q) error estimates to weight fit objective'
        self.input_doc['logI_weighted'] = 'flag for evaluating the fit objective on log(I)'
        self.input_doc['q_range'] = 'lower and upper q-limits for the fit objective'
        self.output_doc['system'] = 'xrsdkit.system.System with optimized parameters'
        self.output_doc['system_dict'] = 'output `system`.to_dict()'
        self.output_doc['q_I_opt'] = 'computed intensity for the optimized system'

    def run(self):
        q_I = self.inputs['q_I']
        sys = self.inputs['system']
        err_wtd = self.inputs['error_weighted']
        logI_wtd = self.inputs['logI_weighted']
        q_range = self.inputs['q_range']
        if sys is not None:
            if sys.fit_report:
                err_wtd = sys.fit_report['error_weighted']
                logI_wtd = sys.fit_report['logI_weighted']
                q_range = sys.fit_report['q_range']

        sys_opt = run_fit_gui(
        sys,q_I[:,0],q_I[:,1],None,err_wtd,logI_wtd,q_range) 

        I_opt = sys_opt.compute_intensity(q_I[:,0]) 
        q_I_opt = np.array([q_I[:,0],I_opt]).T

        self.outputs['system'] = sys_opt
        self.outputs['system_dict'] = sys_opt.to_dict()
        self.outputs['q_I_opt'] = q_I_opt
        return self.outputs




from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    sharpness_limit=40,
    window_width=10
    )
outputs = OrderedDict(
    q_I_dz=None,
    zmask=None
    )

class EasyZingers1d(Operation):
    """Operation for quickly removing zingers from 1d spectral data.

    For each point (or "pixel") in the spectrum, 
    the neighboring points within a specified window width
    to the right and left are analyzed
    to determine whether or not the point of interest 
    is part of an exceedingly sharp edge.
    Points that exceed the sharpness limit are flagged as zingers,
    and in a second pass these points are subtituted 
    by the average intensity in the same window around the zinger point.

    Let pixel_i be a candidate zinger. The analysis is as follows:
    1. Take window_width pixels on either side of pixel_i.
        Call these pixels_left and pixels_right.

    2. (TODO: finish this)
        
    3. (TODO: finish this) 
        
    4. If either of the results in step (3) is greater than sharpness_limit,
        flag pixel_i as a zinger.

    After all zingers are flagged, they are replaced.
    Going over all pixels again, let pixel_i be a zinger:

    1. For pixel_i, take the average of all pixels_left and pixels_right,
        counting only pixels that are not flagged as zingers. 

    2. Replace I(pixel_i) with this window-average value.

    Note, this will result in unusually "flat" 
    regions where zingers used to be,
    as the local window-averaging effectively hides the noise.
    """

    def __init__(self):
        super(EasyZingers1d, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensities'
        self.input_doc['sharpness_limit'] = 'sharpness limit '\
            'for flagging zingers- turn this down to catch more zingers, '\
            'turn it up to catch fewer zingers'
        self.input_doc['window_width'] = 'number of points '\
            'on either side of a given pixel '\
            'used to evaluate sharpness of the pixel'
        self.output_doc['q_I_dz'] = 'same as input q_I but with zingers removed'
        self.output_doc['zmask'] = 'array of booleans, same shape as q, true if there is a zinger at q, else false'

    def run(self):
        q_I = self.inputs['q_I']
        I_ratio_limit = self.inputs['sharpness_limit'] 
        w = self.inputs['window_width'] 
        q = q_I[:,0]
        I = q_I[:,1]
        I_dz = np.array(I)
        zmask = np.array(np.zeros(len(q)),dtype=bool)
        idx_z = []
        I_z = []
        stop_idx = len(q)-w-1
        test_range = iter(range(w,stop_idx))
        idx = next(test_range)
        while idx < stop_idx-1:
            idx_l = np.array([i for i in np.arange(idx-w,idx+1,1) if not i in idx_z])
            idx_r = np.array([i for i in np.arange(idx,idx+w+1,1) if not i in idx_z])
            Ii_l = np.array(I[idx_l])
            Ii_r = np.array(I[idx_r])
            # subtract a linear background from either side
            #Ii_l = Ii_l - np.polyval(np.polyfit(idx_l[:-1],Ii_l[:-1],1),idx_l)
            #Ii_r = Ii_r - np.polyval(np.polyfit(idx_r[1:],Ii_r[1:],1),idx_r)
            # Subtract an approximate linear background from either side
            q_ratio_l = (q[idx_l[-1]] - q[idx_l[:-1]]) / (q[idx_l[-1]]-q[idx_l[0]])  
            q_ratio_r = (q[idx_r[1:]] - q[idx_r[0]])   / (q[idx_r[-1]]-q[idx_r[0]])  
            Ii_l[:-1] = Ii_l[:-1] - q_ratio_l * (Ii_l[0]-Ii_l[-2]) 
            Ii_r[1:] = Ii_r[1:] - q_ratio_r * (Ii_r[-1]-Ii_r[1]) 
            Istd_l = np.std(Ii_l[:-1]) 
            Istd_r = np.std(Ii_r[1:]) 
            
            #I_ratio_l = Ii_l[-1]/Istd_l
            #I_ratio_r = Ii_r[0]/Istd_r
            #I_ratio_l = (Ii_l[-1]-Ii_l[-2])/Imean_diff_l
            #I_ratio_r = (Ii_r[0]-Ii_r[1])/Imean_diff_r
            if Istd_l and Istd_r:
                I_ratio_l = (Ii_l[-1]-Ii_l[-2])/Istd_l
                I_ratio_r = (Ii_r[0]-Ii_r[1])/Istd_r
                if I_ratio_l > I_ratio_limit or I_ratio_r > I_ratio_limit:
                    self.message_callback('found a zinger: q = {}, I = {}'.format(q[idx],I[idx]))
                    idx_z.append(idx)
                    zmask[idx] = True
                    I_dz[idx] = np.nan
            idx = next(test_range) 
        q_z = [q[i] for i in idx_z]
        I_z = [I[i] for i in idx_z]
        newIvals = np.zeros(len(q_z)) 
        for qi,zi in zip(idx_z,range(len(idx_z))):
            Idzi = I_dz[qi-w:qi+w+1]
            Idzi = Idzi[~np.isnan(Idzi)]
            newIvals[zi] = np.mean(Idzi)
        for i,iq in zip(range(len(idx_z)),idx_z):
            I_dz[iq] = newIvals[i]
        self.outputs['q_I_dz'] = np.array([q,I_dz]).T
        self.outputs['zmask'] = zmask




from __future__ import print_function
from collections import OrderedDict
import copy
from threading import Condition

class Operation(object):
    """Class template for implementing paws operations"""

    def __init__(self,inputs,outputs):
        self.inputs = OrderedDict(copy.deepcopy(inputs))
        self.outputs = OrderedDict(copy.deepcopy(outputs))
        self.input_doc = OrderedDict.fromkeys(self.inputs.keys()) 
        self.output_doc = OrderedDict.fromkeys(self.outputs.keys()) 
        self.message_callback = self.tagged_print 
        self.data_callback = None 

        # this lock and flag is used to stop long-running Operations.
        # another process can obtain the lock and set the stop_flag,
        # and then the long-running operation should check the flag's value
        # to decide whether or not to stop.
        self.stop_lock = Condition()
        self.stop_flag = False

    def __getitem__(self,key):
        return self.__dict__[key]

    def __setitem__(self,key,data):
        self.__dict__[key] = data

    def keys(self):
        #return self.__dict__.keys()
        return['inputs','outputs']

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def run_with(self,**kwargs):
        """Run the Operation with keyword arguments substituted for the inputs.

        Any keyword arguments that match the Operation.inputs keys
        are loaded into the Operation.inputs before calling Operation.run().
        All relevant results are stored in Operation.outputs.
        """
        for k in kwargs.keys():
            if not k in self.inputs:
                raise ValueError('Input {} is not valid for Operation {}'.format(k,type(self).__name__))
        self.inputs.update(kwargs)
        return self.run() 

    def run(self):
        """Run the Operation.

        All input data should be specified in Operation.inputs.
        All relevant results should be stored in Operation.outputs.
        """
        pass

    def stop(self):
        with self.stop_lock:
            self.stop_flag = True

    def build_clone(self):
        """Clone the Operation"""
        new_op = self.clone()
        new_op.inputs = copy.deepcopy(self.inputs)
        new_op.outputs = copy.deepcopy(self.outputs)
        return new_op

    @classmethod
    def clone(cls):
        # NOTE: this assumes the subclass constructor 
        # will have default arguments available
        return cls()

    def set_input(self,input_name,val):
        self.inputs[input_name] = val    

    def description(self):
        """Provide a string describing the Operation."""
        return type(self).__name__+": an Operation for PAWS"




from ....IO.BL15.LEGACY.ReadTimeSeries import ReadTimeSeries 
from .. import PlotTimeSeries as PlotTimeSeries_new

class PlotTimeSeries(PlotTimeSeries_new.PlotTimeSeries):
    
    def __init__(self):
        super(PlotTimeSeries,self).__init__()
        # replace reader with LEGACY reader
        self.add_operation('read',ReadTimeSeries())






import copy

from ....IO.BL15.LEGACY.ReadTimeSeries import ReadTimeSeries 
from .. import IntegrateTimeSeries as IntegrateTimeSeries_new

class IntegrateTimeSeries(IntegrateTimeSeries_new.IntegrateTimeSeries):
    
    def __init__(self):
        super(IntegrateTimeSeries,self).__init__()
        # replace reader with LEGACY reader
        self.add_operation('read',ReadTimeSeries())




from datetime import datetime as dt
import importlib
import os
import re
import string 
from collections import OrderedDict

import yaml

from . import operations
from . import workflows
        
bad_chars = string.punctuation 
bad_chars = bad_chars.replace('_','')
bad_chars = bad_chars.replace('-','')
bad_chars = bad_chars.replace('.','')
space_chars = [' ','\t','\n',os.linesep]

p = os.path.abspath(__file__)
# p = (pawsroot)/paws/pawstools.py

d = os.path.dirname(p)
# d = (pawsroot)/paws/
sourcedir = str(d)

d = os.path.dirname(d)
# d = (pawsroot)/
rootdir = str(d)

# Get the code version from the config.py file.
# Reference version string as pawstools.__version__
with open(os.path.join(sourcedir,'config.py')) as f: 
    exec(f.read())

# TODO: ensure this is valid cross-platform
user_homedir = os.path.expanduser("~")

paws_scratch_dir = os.path.join(user_homedir,'.paws_scratch')
paws_cfg_dir = os.path.join(user_homedir,'.paws_cfg')
if not os.path.exists(paws_cfg_dir):
    os.mkdir(paws_cfg_dir)
if not os.path.exists(paws_scratch_dir):
    os.mkdir(paws_scratch_dir)

class WorkflowAborted(Exception):
    pass

class OperationDisabledError(Exception):
    pass

class WfNameError(Exception):
    pass

class PluginNameError(Exception):
    pass

class PluginLoadError(Exception):
    pass

class OperationLoadError(Exception):
    pass

def dtstr():
    """Return date and time as a string"""
    return dt.strftime(dt.now(),'%Y %m %d, %H:%M:%S')

def timestr():
    """Return time as a string"""
    return dt.strftime(dt.now(),'%H:%M:%S')

def save_file(filename,d):
    """
    Create or replace file indicated by filename,
    as a yaml serialization of dict d.
    """
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()
    
def update_file(filename,d):
    """
    Save the items in dict d into filename,
    without removing members not included in d.
    """
    if os.path.exists(filename):
        f_old = open(filename,'r')
        d_old = yaml.load(f_old)
        f_old.close()
        d_old.update(d)
        d = d_old
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()

class DictTree(object):
    """A data structure for tree-like storage.

    A DictTree has a root (an ordered dictionary), 
    which is extended by embedding other objects 
    that are amenable to tree storage.
    Fetches items by keys (strings),
    which are sequences 
    of parent item keys(), connected by '.'s.

    Child items (end nodes of the tree)
    can be anything.
    Parent items, in order to index their children,
    must be either lists, dicts, or objects implementing
    keys(), __getitem__(key) and __setitem__(key,value).
    """

    def __init__(self,data={}):
        super(DictTree,self).__init__()
        self._root = OrderedDict()
        if isinstance(data,dict):
            self._root = OrderedDict(data)

    def __getitem__(self,key):
        return self.get_data(key)

    def __setitem__(self,key,val):
        self.set_data(key,val)

    def root_keys(self):
        return self._root.keys()

    def keys(self):
        return self.subkeys()

    def subkeys(self,root_key=''):
        if not root_key:
            itm = self._root
        else:
            itm = self.get_data(root_key)
        itm_keys = []
        try:
            itm_keys = itm.keys()
        except:
            if isinstance(itm,list): 
                itm_keys = [str(i) for i in range(len(itm))] 
            else:
                # no itm_keys
                pass
        prefix = root_key
        if root_key: prefix += '.'
        sk = [prefix+s for s in itm_keys]
        for k in itm_keys:
            next_keys = self.subkeys(prefix+k) 
            if any(next_keys):
                sk.extend(next_keys)
        return sk

    @ staticmethod
    def parent_key(key):
        # TODO: handle the possibility of subkeys containing periods
        if '.' in key:
            return key[:key.rfind('.')]
        else:
            return ''

    def delete_data(self,key=''):
        """Attempts to de-reference the given key"""
        parent_itm = self._root
        if '.' in key:
            parent_itm = self.get_data(self.parent_key(key))
        itm_key = key.split('.')[-1]
        if itm_key:
            try: 
                # parent implements __getitem__()?
                parent_itm.pop(itm_key)
            except:
                # parent item is list? 
                parent_itm.pop(int(itm_key))

    def set_data(self,key='',val=None):
        """Sets the data at the given key."""
        parent_itm = self._root
        if '.' in key:
            parent_itm = self.get_data(self.parent_key(key))
        itm_key = key.split('.')[-1]
        if itm_key:
            try: 
                parent_itm[itm_key] = val
            except:
                try: 
                    parent_itm[int(itm_key)] = val # list case
                except:
                    parent_itm.append(val) # append to list case

    def get_data(self,key=''):
        """Returns the data at the given key."""
        path = key.split('.')
        itm = self._root 
        for ik,k in enumerate(path):
            child_found = False
            try: 
                itm = itm[k]
                child_found = True
            except:
                try: 
                    itm = itm[int(k)]
                    child_found = True
                except:
                    longer_key = k
                    for kk in path[ik+1:]:
                        longer_key += '.'
                        try: 
                            itm = itm[longer_key]
                            child_found = True
                        except: 
                            pass
                        longer_key += kk
                        try: 
                            itm = itm[longer_key]
                            child_found = True
                        except: 
                            pass
            if not child_found:
                raise KeyError(key)
        return itm

    def is_key_valid(self,key):
        """Check key validity.

        Keys may contain upper case letters, lower case letters, 
        numbers, dashes (-), and underscores (_) and period (.) marks. 
        Any whitespace or any character in the string.punctuation library
        (other than -, _, or .) results in an invalid key.
        """
        if not key or any(map(lambda s: s in key,space_chars))\
            or any(map(lambda s: s in key,bad_chars)):
            return False 
        return True 

    def key_error_message(self,key):
        """Provide a human-readable error message for bad keys."""
        if not key:
            return 'key is blank.'
        elif any(map(lambda s: s in key,space_chars)):
            return '"{}" contains whitespace.'.format(key)
        elif any(map(lambda s: s in key,bad_chars)):
            return '"{}" contains special characters.'.format(key)

    def print_tree(self,root_key='',offset=''):
        """Print the tree content."""
        itm = self._root
        if root_key:
            itm = self.get_data(root_key)
        tstr = os.linesep 
        try:    #if isinstance(itm,dict):
            for k in itm.keys():
                x_str = self.print_tree(root_key+'.'+k,offset+'    ')
                tstr = tstr+offset+'{}: {}'.format(k,x_str)+os.linesep
        except:
            try:    #elif isinstance(itm,list):
                for i,x in enumerate(itm):
                    x_str = self.print_tree(root_key+'.'+str(i),offset+'    ')
                    tstr = tstr+offset+'{}: {}'.format(i,x_str)+os.linesep
            except:
                return '{}'.format(itm)
        return tstr





from collections import OrderedDict
import copy

import numpy as np

from ..Operation import Operation

inputs = OrderedDict(
    batch_outputs={},
    x_key=None,
    x_sort_flag=True,
    x_shift_flag=False, 
    lower_index=None,
    upper_index=None,
    index_step=1) 
outputs = OrderedDict(
    x_list = None,
    x_array = None,
    sorted_outputs = {}) 

class SortBatch(Operation):
    """
    Harvest sorted lists from a batch output (a dict of unsorted lists). 
    Takes a batch output, and a key for the values on which to sort.
    """

    def __init__(self):
        super(SortBatch,self).__init__(inputs,outputs)        
        self.input_doc['batch_outputs'] = 'dict of lists, where all lists are in corresponding order'
        self.input_doc['x_key'] = 'key for x data from batch_outputs'
        self.input_doc['x_sort_flag'] = 'if True, sort data for increasing x' 
        self.input_doc['x_shift_flag'] = 'if True, shift x data so that its minimum value is zero' 
        self.input_doc['lower_index'] = 'optional list slice lower limit, inclusive'
        self.input_doc['upper_index'] = 'optional list slice upper limit, exclusive'
        self.input_doc['index_step'] = 'optional number of indices to skip between sorted outputs'
        self.output_doc['x_list'] = 'list of values of x (sorted)'
        self.output_doc['x_array'] = 'array of values of x (sorted)'
        self.output_doc['sorted_outputs'] = 'dict similar to inputs, but sorted on `x_key`'

    def run(self):
        b_out = self.inputs['batch_outputs']

        kx = self.inputs['x_key']
        sortflag = self.inputs['x_shift_flag']
        shiftflag = self.inputs['x_shift_flag']
        skipidx = self.inputs['index_step']
        lidx = self.inputs['lower_index']
        uidx = self.inputs['upper_index']

        x_list = b_out[kx]
        n_batch_outputs = len(x_list) 
        if shiftflag or sortflag or uidx is not None or lidx is not None:
            xa = np.array(x_list)
            if shiftflag:
                xmin = min(x_list)
                xa = xa - xmin
            ix = np.arange(len(xa))
            if sortflag:
                ix = np.argsort(xa)
            xa = xa[ix]
            if lidx is not None:
                ix = ix[lidx:]
                xa = xa[lidx:]
                if uidx is not None:
                    uidx = uidx-lidx
            if uidx is not None:
                ix = ix[:uidx]
                xa = xa[:uidx]
            if skipidx > 1:
                ix = ix[::skipidx]
                xa = xa[::skipidx]
            self.outputs['x_array'] = xa
            self.outputs['x_list'] = [x_list[int(ii)] for ii in ix]

        self.message_callback('sorting {} on key: "{}"'.format(list(b_out.keys()),kx))
        s_out = self.outputs['sorted_outputs']
        for y_key in b_out.keys():
            y_list = copy.deepcopy(b_out[y_key])

            if shiftflag or sortflag or uidx is not None or lidx is not None:
                # only sort y_list if it contains a full batch of outputs
                if len(y_list) == n_batch_outputs:
                    y_list = [y_list[int(ii)] for ii in ix]
        
            s_out[y_key] = y_list

        return self.outputs




import os
import pkgutil

op_modules = []
def load_ops_from_path(path_,pkg,cat_root=''):
    ops = []
    cats = []
    mods = pkgutil.iter_modules(path_)
    mods = [mod for mod in mods if mod[1] not in ['__init__','Operation','OpManager','optools']]
    for modloader, modname, ispkg in mods:
        if ispkg:
            pkg_path = [os.path.join(path_[0],modname)]
            subcat_root = modname
            if cat_root:
                subcat_root = cat_root+'.'+modname
            pkg_ops, pkg_cats = load_ops_from_path(pkg_path,pkg+'.'+modname,subcat_root)
            pkg_ops = [op for op in pkg_ops if not op in ops]
            pkg_cats = [cat for cat in pkg_cats if not cat in cats]
            ops = ops + pkg_ops
            cats = cats + pkg_cats
        else:
            # assume that there is an Operation in this module
            # whose class name is the same as the module name.
            if not cat_root in cats:
                cats.append(cat_root)
            ops.append( (cat_root,modname) )
            op_modules.append(cat_root+'.'+modname)
    return ops, cats

cat_op_list, cat_list = load_ops_from_path(__path__,__name__)




from __future__ import print_function
from collections import OrderedDict
import glob
import os
from threading import Condition

import paramiko

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    username=None,
    hostname=None,
    password=None,
    port=22,
    private_key_file=None)

# TODO: handle non-unix hosts

class SSHClient(PawsPlugin):
    """An SSH Client PawsPlugin"""

    def __init__(self):
        super(SSHClient,self).__init__(content)
        self.content_doc['username'] = 'remote user name'
        self.content_doc['hostname'] = 'remote host name'
        self.content_doc['password'] = 'user password on remote host'
        self.content_doc['port'] = 'port for ssh connection, default 22'
        self.content_doc['private_key_file'] = 'path to private RSA key file'
        self.sshcl = None
        self.sftpcl = None
        self.thread_blocking = True
        #self.copy_lock = Condition()
        self.transport_lock = Condition()

    def start(self):
        super(SSHClient,self).start()
        # block until notification that the clone is running 
        with self.thread_clone.running_lock:
            self.thread_clone.running_lock.wait()

    def run(self):
        # this method is run by self.thread_clone 
        username = self.content['username']
        hostname = self.content['hostname']
        pw = self.content['password']
        port = self.content['port']
        pkey_file = self.content['private_key_file']
        with self.transport_lock:
            self.sshcl = paramiko.SSHClient()
            self.sshcl.load_system_host_keys()
            if pkey_file is not None:
                k = paramiko.RSAKey.from_private_key_file(pkey_file)
                self.sshcl.connect(hostname, port, pkey=k)
            else:
                self.sshcl.connect(hostname, port, username, pw)
            self.sftpcl = paramiko.SFTPClient.from_transport(self.sshcl.get_transport())
        if self.verbose: self.message_callback('SSH client connected to {} port {}'.format(hostname,port))
        self.run_notify()

    def stop(self):
        super(SSHClient,self).stop() 
        if self.thread_clone:
            self.thread_clone.close_ssh()

    def close_ssh(self):
        with self.transport_lock:
            if self.sshcl:
                self.sshcl.close()
                self.sshcl = None

    def copy_file(self,remote_path,local_path):
        if self.verbose: self.message_callback('attempting to copy {}'.format(remote_path))
        try_again = True
        n_tries = 0
        with self.thread_clone.transport_lock:
            while try_again:
                try:
                    self.thread_clone.sftpcl.get(remote_path,local_path)
                    try_again = False
                except:
                    n_tries += 1
                    if n_tries > 10:
                        self.message_callback('failed to copy {}'.format(remote_path))
                        try_again = False
                    else:
                        self.message_callback('attempt {} to copy {}'.format(n_tries,remote_path))
        if self.verbose: self.message_callback('finished copying to {}'.format(local_path))

#    def sync_files(self,remote_dir,regex,local_dir,overwrite=False):
#        rrx = os.path.join(remote_dir,regex)
#        if self.verbose: self.message_callback('attempting copy from {}'.format(rrx))
#        stdin, stdout, stderr = self.sshcl.exec_command('ls {}'.format(rrx))
#        rdir_list = stdout.read().strip().split(os.linesep.encode())
#        ldir_list = glob.glob(os.path.join(local_dir,'*'))
#        ldir_file_list = [os.path.split(fpath)[1] for fpath in ldir_list]
#        #if self.verbose: self.message_callback('current files: {}'.format(rdir_list))
#        #print(rdir_list)
#        #with self.directory_lock:
#        for fpath in rdir_list:
#            fn = os.path.split(fpath.decode())[1] 
#            if fn in ldir_file_list:
#                if overwrite:
#                    if self.verbose: self.message_callback('overwriting file: {}'.format(fn))
#                    self.sftpcl.get(os.path.join(remote_dir,fn),os.path.join(local_dir,fn))
#                else:
#                    if self.verbose: self.message_callback('{} exists locally: skipping copy'.format(fn))
#            else:
#                #self.remote_dir.append(fn) 
#                #with self.proxy.directory_lock:
#                #    # update the proxy's directory
#                #    self.proxy.remote_dir.append(fn)
#                if self.verbose: self.message_callback('copying new file: {}'.format(fn))
#                self.sftpcl.get(os.path.join(remote_dir,fn),os.path.join(local_dir,fn))
#        self.notify_locks([self.copy_lock])
#



from collections import OrderedDict

import numpy as np

from ..Operation import Operation

inputs = OrderedDict(size=100)
outputs = OrderedDict(array=None)

class NoiseArray(Operation):
    """Creates and outputs a square array of noise"""

    def __init__(self):
        super(NoiseArray,self).__init__(inputs,outputs) 
        self.input_doc['size'] = 'dimension of output array'
        self.output_doc['array'] = 'an array of noise'
    
    def run(self):
        s = self.inputs['size']
        self.outputs['array'] = np.random.rand(s,s) 
        return self.outputs




import os

from paws.workflows.FITTING.XRSDFit import XRSDFit 
from paws.operations.PROCESSING.FITTING.XRSDFitGUI import XRSDFitGUI as XRSDFitGUI_op

class XRSDFitGUI(XRSDFit):

    def __init__(self):
        super(XRSDFitGUI,self).__init__()
        self.add_operations(fit = XRSDFitGUI_op())




import copy

from paws.workflows.FITTING.BL15 import TimeSeriesXRSDFit
from paws.workflows.FITTING import XRSDFitGUI

class TimeSeriesXRSDFitGUI(TimeSeriesXRSDFit.TimeSeriesXRSDFit):

    def __init__(self):
        super(TimeSeriesXRSDFitGUI,self).__init__()
        self.add_operation('fit',XRSDFitGUI.XRSDFitGUI())




from collections import OrderedDict

from xrsdkit.tools import piftools

from ...Operation import Operation

inputs=OrderedDict(pif=None)
outputs=OrderedDict(
    experiment_id=None,
    t_utc=None,
    q_I=None,
    temperature=None,
    features=None,
    system=None)

class UnpackXRSDPIF(Operation):
    """Unpack a PIF record that was generated by xrsdkit"""

    def __init__(self):
        super(UnpackXRSDPIF,self).__init__(inputs,outputs)
        self.input_doc['pif'] = 'pif object to be unpacked'
        self.output_doc['experiment_id'] = 'string experiment id'
        self.output_doc['t_utc'] = 'time in seconds utc'
        self.output_doc['q_I'] = 'n-by-2 array of q values and measured saxs intensities'
        self.output_doc['temperature'] = 'temperature in degrees C'
        self.output_doc['features'] = 'dict of numerical features of `q_I`'
        self.output_doc['system'] = 'xrsdkit.system.System object'

    def run(self):
        pp = self.inputs['pif']

        _uid, sys, q_I, expt_id, t_utc, T_C, _src_wl, feats, _cls_labels, _reg_labels = piftools.unpack_pif(pp)

        self.outputs['experiment_id'] = expt_id 
        self.outputs['t_utc'] = t_utc 
        self.outputs['q_I'] = q_I 
        self.outputs['temperature'] = T_C 
        self.outputs['features'] = feats 
        self.outputs['system'] = sys 




from collections import OrderedDict
import os
import time

import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    flow_reactor=None)
outputs=OrderedDict(
    recipe={},
    header={})
 
class ReadFlowReactor(Operation):
    """Read and package the current state of a FlowReactor Plugin."""

    def __init__(self):
        super(ReadFlowReactor,self).__init__(inputs,outputs)
        self.input_doc['flow_reactor'] = 'FlowReactor plugin'
        self.output_doc['recipe'] = 'FlowReactor recipe data'
        self.output_doc['header'] = 'dict of data related to the recipe'

    def run(self):
        fr = self.inputs['flow_reactor'] 
        flag,statstr,hdr = fr.check_status()
        rcp = {}
        rcp['T_set'] = float(hdr['T_set_A'])
        #rcp['T_read'] = float(hdr['T_read_A'])
        setpt_keys = [k for k in hdr.keys() if 'setpoint' in k]
        fr_tot = np.sum([hdr[k] for k in setpt_keys])
        rcp['flowrate'] = float(fr_tot)
        rcp['solvent'] = fr.solvent_name
        for k in setpt_keys:
            if not k == fr.solvent_name+'_setpoint' in k:
                rg_name = k[:k.find('_setpoint')]
                fr_rg = float(hdr[k])
                rcp[rg_name+'_fraction'] = float(fr_rg/fr_tot)
        #self.message_callback('measured recipe: {}'.format(rcp))
        #self.message_callback('header data: {}'.format(hdr))
        self.outputs['recipe'] = rcp
        self.outputs['header'] = hdr 
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import socket 
import copy
from threading import Condition

from .PawsPlugin import PawsPlugin
from .. import pawstools

content = OrderedDict(
    host=None,
    port=None,
    channels={},
    timer=None)

class CryoConController(PawsPlugin):

    def __init__(self):
        super(CryoConController,self).__init__(content)
        self.thread_blocking = True
        self.content_doc['host'] = 'string representing host name or IP address assigned to the CryoCon'
        self.content_doc['port'] = 'integer port number for CryoCon connection' 
        self.content_doc['channels'] = 'dict mapping channel selection (strings "A"-"D") '\
            'to control loop indices (integers 1-4 or None): all channels will be read, '\
            'and any with integer control loops will be controlled. '
        self.content_doc['timer'] = 'Timer plugin for triggering client activities' 
        self.socket_lock = Condition()
        self.sock = None
        self.state_lock = Condition()
        self.state = {} 

    def description(self):
        desc = 'CryoConController Plugin: '\
            'This is a TCP Client used to communicate with a CryoCon. '\
            'Startup requires a host name and a port number, '\
            'which should match the network configuration of the CryoCon device.'
        return desc

    def start(self):
        super(CryoConController,self).start()

    def run(self):
        # executed by self.thread_clone in its own thread
        hst = self.content['host'] 
        prt = self.content['port'] 
        with self.socket_lock:
            self.sock = socket.create_connection((hst,prt)) 
        tmr = self.content['timer'] 
        self.run_cmd('*idn?')
        keep_going = True

        self.update_status()

        while keep_going: 
            with tmr.dt_lock:
                tmr.dt_lock.wait()
            #while self.commands.qsize() > 0:
            #    with self.command_lock:
            #        cmd = self.commands.get()
            #        self.commands.task_done()
            #
            #    self.run_cmd(cmd)
            self.update_status()
            with tmr.running_lock:
                if not tmr.running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        if self.verbose: self.message_callback('FINISHED')
        self.run_cmd('stop')
        with self.socket_lock:
            self.sock.close()

    def read_status(self):
        for chan in self.content['channels'].keys():
            resp = self.run_cmd('input {}:temp?'.format(chan))
            with self.state_lock:
                self.state['T_read_{}'.format(chan)] = float(resp)

    def update_status(self):
        with self.state_lock:
            self.read_status()
        with self.proxy.state_lock:
            self.proxy.state = copy.deepcopy(self.state)
        #if self.verbose: self.message_callback('T_read: {}'.format(self.state['T_read']))

    def run_cmd(self,cmd):
        with self.content['timer'].dt_lock:
            t_now = float(self.content['timer'].dt_utc())
        with self.socket_lock:
            self.send_line(cmd)
            resp = self.receive_line()
        if self.verbose: self.message_callback('{}: {}'.format(t_now,cmd+' '+resp))
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd+' '+resp)
        return resp

    def send_line(self, line):
        self.sock.sendall(bytearray((line+'\n').encode('utf-8')))

    def receive_line(self):
        bfr = bytearray(b' ' * 1024) 
        ln = self.sock.recv_into(bfr) 
        bfr = bfr.strip().decode()
        return bfr

    def set_units(self,channel,unit_str):
        cmd = 'input {}:units {}'.format(channel,unit_str)
        self.thread_clone.run_cmd(cmd)
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put(cmd)

    def loop_setup(self,channel,control_type='PID'):
        cmd = 'loop {}:source {};type {}'.format(self.content['channels'][channel],channel,control_type)
        self.thread_clone.run_cmd(cmd)
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put(cmd)

    def set_temperature(self,channel,T_set):
        cmd = 'loop {}:setpt {}'.format(self.content['channels'][channel],T_set)
        self.thread_clone.run_cmd(cmd)
        with self.thread_clone.state_lock:
            self.thread_clone.state['T_set_{}'.format(channel)] = T_set 
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put(cmd)

    def set_ramp_rate(self,channel,T_ramp):
        cmd = 'loop {}:rate {}'.format(self.content['channels'][channel],T_ramp)
        self.thread_clone.run_cmd(cmd)
        with self.thread_clone.state_lock:
            self.thread_clone.state['T_ramp_{}'.format(channel)] = T_ramp 
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put(cmd)

    def start_control(self):
        self.thread_clone.run_cmd('control')
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put('control')
 
    def stop_control(self):
        with self.thread_clone.running_lock:
            self.thread_clone.run_cmd('stop')
        #with self.thread_clone.command_lock:
        #    self.thread_clone.commands.put('stop')
 



from collections import OrderedDict

from ..Operation import Operation

inputs = OrderedDict(data=None)
outputs = OrderedDict()

class Print(Operation):
    """Operation for printing data"""

    def __init__(self):
        super(Print,self).__init__(inputs,outputs) 
        self.input_doc['data'] = 'an object to print'
        
    def run(self):
        self.message_callback(self.inputs['data'])
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import socket 
from threading import Condition
import time

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    host=None,
    port=None)

class SpecInfoClient(PawsPlugin):

    def __init__(self):
        super(SpecInfoClient,self).__init__(content)
        self.content_doc['host'] = 'string representing host name or IP address'
        self.content_doc['port'] = 'integer port number where SpecInfoServer listens' 
        self.socket_lock = Condition()
        self.sock = None
        self.thread_blocking = True

    def description(self):
        desc = 'SpecInfoClient Plugin: '\
            'This is a TCP Client used to communicate with SpecInfoServer. '\
            'Startup requires a host name and a port number, '\
            'where SpecInfoServer should be listening.'
        return desc

    def start(self):
        super(SpecInfoClient,self).start()

    def run(self):
        hst = self.content['host'] 
        prt = self.content['port'] 
        with self.socket_lock:
            self.sock = socket.create_connection((hst,prt)) 

        self.take_control()

    def stop(self):
        super(SpecInfoClient,self).stop() 
        if self.thread_clone:
            self.thread_clone.close_socket()

    def close_socket(self):
        self.sock.close()

    def run_cmd(self,cmd):
        resp = ''
        while resp in ['','spec is busy!']:
            with self.socket_lock:
                self.send_line(cmd)
                resp = self.receive_line()
        with self.proxy.history_lock:
            self.proxy.add_to_history(self.get_date_time(),cmd+' '+resp)
        if self.verbose: self.message_callback(cmd+' '+resp)
        return resp

    def send_line(self, line):
        self.sock.sendall(bytearray(line.encode('utf-8')))

    def receive_line(self):
        bfr = bytearray(b' ' * 1024) 
        ln = self.sock.recv_into(bfr) 
        bfr = bfr.strip().decode()
        return bfr
    
    def take_control(self):
        resp = self.run_cmd('!rqc')
        while not resp == 'client in control.':
            time.sleep(0.1)
            resp = self.run_cmd('!rqc')

    def mar_expose(self,filename,exposure_time):
        self.thread_clone.run_cmd('!cmd mar netroot {}'.format(filename))
        self.thread_clone.run_cmd('!cmd mar collect {}'.format(exposure_time))
        sleep_time = exposure_time+5
        self.message_callback('waiting {} seconds for {}-second exposure'.format(sleep_time,exposure_time))
        time.sleep(sleep_time)

#    def enable_cryocon(self):
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd ctemp_enable')
#            self.thread_clone.commands.put('!cmd ctemp_ctrl_on')
#
#    def set_cryocon(self,temperature,ramp=None):
#        with self.thread_clone.command_lock:
#            if ramp is not None:
#                self.thread_clone.commands.put('!cmd ctemp_ramp_on')
#                self.thread_clone.commands.put('!cmd csetramp {}'.format(ramp))
#            self.thread_clone.commands.put('!cmd csettemp {}'.format(temperature))
#
#    def stop_cryocon(self):
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd ctemp_ctrl_off')
#            self.thread_clone.commands.put('!cmd ctemp_disable')
#
#    def read_cryocon(self):
#        # !cmd cmeasuretemp     -> reads temperature, saves as CYRO_DEGC
#        # !cmd CRYO_DEGC        -> query the CRYO_DEGC variable
#        # ?res                  -> gets result of CRYO_DEGC query
#        # TODO: this should block until it has a result,
#        # and then it should return that result
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd cmeasuretemp')
#
#    def mar_enable(self):
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd mar_enable')
# 
#    def mar_disable(self):
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd mar_disable')
# 
#    def run_loopscan(self,mroot,n_points,exp_time,block=False):
#        # !cmd loopscan n_points exposure_time sleep_time
#        with self.thread_clone.command_lock:
#            self.thread_clone.commands.put('!cmd mar netroot {}'.format(mroot))
#            self.thread_clone.commands.put('!cmd loopscan {} {}'.format(n_points,exp_time))
#        if block:
#            # wait for self.thread_clone.loopscan_lock
#            if self.verbose: self.message_callback('blocking until loopscan finishes')
#            with self.thread_clone.loopscan_lock:
#                self.thread_clone.loopscan_lock.wait()
#            if self.verbose: self.message_callback('loopscan finished')
#
#------------------------------------------------------------------------------

#    p.addCommand("!cmd slacx_mar_data_path = 'my_mar_data'")
#    p.addCommand("!cmd slacx_pd_filename = 'my_pd_filename'")
#    p.addCommand("!cmd slacx_loopscan_npoints = 2")
#    p.addCommand("!cmd slacx_loopscan_counting_time = 2")
#    p.addCommand("!cmd runme")
#    p.addCommand("?sta")
#
# !rqc                  -> requests control
# !cmd mar_enable       -> enables mar det as a counter
# !cmd pd enable        -> enables pilatus as a counter

# Loop Scanning
# !cmd loopscan n_points exposure_time sleep_time





from collections import OrderedDict
import os

import pypif.obj as pifobj
import yaml
import numpy as np
from xrsdkit.tools import piftools

from ...Operation import Operation

inputs=OrderedDict(
    recipe_data={},
    q_I=None,
    system=None,
    additional_properties={}
    )
outputs=OrderedDict(pif=None)

class FlowSynthesisPIF(Operation):
    """Build a PIF record for a flow reactor synthesis experiment"""

    def __init__(self):
        super(FlowSynthesisPIF,self).__init__(inputs,outputs)
        self.input_doc['recipe_data'] = 'Header containing recipe data'
        self.input_doc['q_I'] = 'n-by-2 array of scattering vectors and intensities'
        self.input_doc['system'] = 'xrsdkit.system.System describing '\
            'the populations and physical parameters of the material system'
        self.input_doc['additional_properties'] = 'dict of additional properties '\
            'to add to the PIF (each key is a property name, each value is a value)'
        self.output_doc['pif'] = 'pif object representing the synthesis experiment'

    def run(self):
        rcp = self.inputs['recipe_data']
        q_I = self.inputs['q_I']
        sys = self.inputs['system']
        uid = sys.sample_metadata['sample_id']
        if not uid:
            uid = 'tmp'
            if sys.sample_metadata['experiment_id']:
                uid = expt_id 
            if sys.sample_metadata['t_utc']:
                uid = uid+'_'+str(int(t_utc))

        self.message_callback('building PIF (sample_id: {})'.format(sys.sample_metadata['sample_id']))
        csys = piftools.make_pif(sys,q_I)

        for nm,val in rcp.items():
            prop = pifobj.Property(nm,val)
            prop.tags = ['flow reactor recipe data']
            prop.dataType = 'EXPERIMENTAL'
            csys.properties.append(prop)

        for nm,val in self.inputs['additional_properties'].items():
            prop = pifobj.Property(nm,val)
            csys.properties.append(prop)

        self.outputs['pif'] = csys
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
from .PawsPlugin import PawsPlugin
import datetime
import copy
import time
from threading import Condition

import tzlocal
import numpy as np

content = OrderedDict(
    dt=1.,
    t_max=None,
    dt_now=None,
    utc_now=None)

class Timer(PawsPlugin):
    """A Paws plugin for marking the passage of time"""

    def __init__(self):
        super(Timer,self).__init__(content)
        self.content_doc['dt'] = 'Wait time between signals, in seconds'
        self.content_doc['t_max'] = 'Seconds after which the timer stops itself'
        self.content_doc['dt_now'] = 'Seconds from Timer.start() to the most recent Timer.dt_now()'
        self.content_doc['utc_now'] = 'Time in seconds utc at the most recent Timer.dt_now()'
        # A lock for the "tick" trigger 
        self.dt_lock = Condition()
        self.t0 = None 
        self.tz = tzlocal.get_localzone()
        self.ep = datetime.datetime.fromtimestamp(0,self.tz)
        self.t0_utc = None 
        self.thread_blocking = True

    def start(self):
        self.tz = tzlocal.get_localzone()
        self.ep = datetime.datetime.fromtimestamp(0,self.tz)
        self.t0 = datetime.datetime.now(self.tz)
        self.t0_utc = (self.t0-self.ep).total_seconds()
        super(Timer,self).start()

    def run(self):
        # self.thread_clone runs this method in its own thread.
        if self.verbose: self.message_callback('timer STARTED')
        #with self.proxy.dt_lock:
        dt = self.content['dt'] 
        t_max = self.content['t_max']
        keep_going = True
        # set zero-time point 
        self.t0 = copy.deepcopy(self.proxy.t0)
        self.t0_utc = copy.deepcopy(self.proxy.t0_utc)
        dt_now = self.dt_utc()
        self.proxy.dt_notify()
        #dt_err = np.mod(dt_now,dt)
        while dt_now < t_max and keep_going:
            if self.verbose: self.message_callback('tick: {}/{}'.format(dt_now,t_max))
            # attempt to nail the next dt point:
            t_rem = dt-np.mod(self.dt_utc(),dt)
            #time.sleep(t_rem-dt_err)
            time.sleep(t_rem)
            # acquire proxy dt lock, take time point, dt_notify, release
            #with self.proxy.dt_lock:
            dt_now = self.dt_utc()
            self.proxy.dt_notify()
            #dt_err = np.mod(dt_now,dt)
            # TODO: implement feedback to hone the timer
            # acquire proxy running_lock, check if we are still running
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
                if self.verbose and not keep_going: self.message_callback('timer STOPPED')
        # take a few extra ticks for listening plugins to stop themselves?
        if self.verbose: self.message_callback('Ticking a few more times before stopping')
        for itick in range(3):
            time.sleep(dt)
            self.proxy.dt_notify()
        with self.proxy.running_lock:
            self.proxy.stop()
        # now that the plugin is stopped, notify any other plugins that are waiting on dt_lock
        #with self.proxy.dt_lock:
        self.proxy.dt_notify()
        if self.verbose: self.message_callback('Timer FINISHED')

    def dt_notify(self):
        with self.dt_lock:
            if int(self.py_version) == 2:
                self.dt_lock.notifyAll()
            else:
                self.dt_lock.notify_all()

    def t_utc(self):
        t = datetime.datetime.now(self.tz)
        return float((t-self.ep).total_seconds())

    def time_as_string(self):
        return str(datetime.datetime.now(self.tz))

    def dt_utc(self):
        tnow = self.t_utc()
        dt = float(tnow-self.t0_utc)
        self.content['utc_now'] = tnow
        self.content['dt_now'] = dt
        return dt

    def description(self):
        desc = str('Timer Plugin for Paws: '\
            'Signals the passage of time, '\
            'as a convenience for other plugins '\
            'who depend on a clock to trigger their actions')
        return desc





from collections import OrderedDict
import os
import time

from ...Operation import Operation

inputs=OrderedDict(ppump_controllers={})
outputs=OrderedDict()
        
class TarePPumps(Operation):
    """Tare an array of P-pump controllers."""

    def __init__(self):
        super(TarePPumps,self).__init__(inputs,outputs)
        self.input_doc['ppump_controllers'] = 'dict of MitosPPumpController plugins'

    def run(self):
        ppcs = self.inputs['ppump_controllers'] 
        for ppc_name,ppc in ppcs.items():
            ppc.tare()




import os

from paws.workflows.IO.BL15.LEGACY.ReadBatch import ReadBatch
from paws import pawstools

tdd = os.path.join(pawstools.rootdir,'tests','test_data')

def test():
    batchread_wf = ReadBatch()
    batchread_wf.run_with(
        time_key = 't_utc',
        header_dir = os.path.join(tdd,'headers','legacy'),
        header_regex = '*.txt',
        image_dir = os.path.join(tdd,'images'),
        q_I_dir = os.path.join(tdd,'data'),
        system_dir = os.path.join(tdd,'systems')
        )
    return True



import os
import copy
from collections import OrderedDict

import numpy as np

from ...Workflow import Workflow 
from ...IO.BL15.ReadTimeSeries import ReadTimeSeries 
from ....operations.PACKAGING.PIF.FlowSynthesisPIF import FlowSynthesisPIF
from ....operations.IO.CITRINATION.UploadPIF import UploadPIF

inputs = OrderedDict(
    citrination_client = None,
    experiment_id = 'test',
    source_wavelength = None,
    header_dir = None,
    header_suffix = None,
    header_regex = None,
    data_dir = None,
    q_I_suffix = None,
    system_suffix = None,
    #design_goals = {},
    dataset_id = None,
    upload_pif = False,
    keep_pif_json = True
    )

outputs = OrderedDict()

class TimeSeriesFlowReactorPIF(Workflow):
    
    def __init__(self):
        super(TimeSeriesFlowReactorPIF,self).__init__(inputs,outputs)
        self.add_operations(
            read = ReadTimeSeries(),
            make_pif = FlowSynthesisPIF(),
            upload_pif = UploadPIF()
            )

    def run(self):
        expt_id = self.inputs['experiment_id']
        src_wl = self.inputs['source_wavelength']
        # READ THE TIMESERIES 
        self.operations['read'].operations['read_batch'].operations['read'].disable_ops('read_image')
        ts_data = self.operations['read'].run_with(
            header_dir = self.inputs['header_dir'],
            q_I_dir = self.inputs['data_dir'], 
            system_dir = self.inputs['data_dir'], 
            header_regex = self.inputs['header_regex'],
            header_suffix = self.inputs['header_suffix'],
            q_I_suffix = self.inputs['q_I_suffix'],
            system_suffix = self.inputs['system_suffix']
            )
        # CREATE TIMESERIES OF PIFS
        all_pifs = []
        ts = ts_data['time']
        hdrs = ts_data['header_data']
        syss = ts_data['system']
        q_Is = ts_data['q_I']
        for ii, t_utc, hdr_data, sys, q_I in zip(range(len(ts)),ts,hdrs,syss,q_Is): 
            pif_outputs = self.operations['make_pif'].run_with(
                experiment_id = expt_id+'_{}'.format(ii),
                source_wavelength = src_wl,
                t_utc = t_utc,
                header_data = hdr_data,
                #design_goals = self.inputs['design_goals'],
                q_I = q_I, 
                system = sys 
                )
            all_pifs.append(pif_outputs['pif'])
        # SAVE AND UPLOAD THE PIFS
        pif_json_path = os.path.join(self.inputs['data_dir'],expt_id+'_pif.json') 
        self.operations['upload_pif'].run_with(
            pif = all_pifs,
            citrination_client = self.inputs['citrination_client'],
            dsid = self.inputs['dataset_id'],
            json_path = pif_json_path,
            keep_json = self.inputs['keep_pif_json'],
            upload_flag = self.inputs['upload_pif']
            )
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import os

import numpy as np
import pyFAI

from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(poni_dict=None)

class NikaToPONI(Operation):
    """
    Converts Nika calibration output 
    (saved in a text file)
    to a dict of PyFAI PONI parameters,
    by first converting from Nika to Fit2D,
    then using a pyFAI.AzimuthalIntegrator 
    to convert from Fit2D to PONI format.

    WARNING: the map from Nika's horizontal and vertical tilts
    to Fit2D's tilt and tiltPlanRotation
    has not yet been verified by the developers. 
    Use this operation with nonzero tilts at your own risk.
 
    Input a text file expressing results of Nika automated calibration.
    Output a dict of pyFAI PONI calibration parameters.
    Format of text file for Nika output is expected to be:
    sample_to_CCD_mm=____
    pixel_size_x_mm=____
    pixel_size_y_mm=____
    beam_center_x_pix=____
    beam_center_y_pix=____
    horizontal_tilt_deg=____
    vertical_tilt_deg=____ 
    wavelength_A=____ 
    """
    
    def __init__(self):
        super(NikaToPONI,self).__init__(inputs,outputs)
        self.input_doc['file_path'] = 'text file expressing nika automated calibration results- '\
            'see documentation of this operation class for the expected format of this file'
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters, as found in a .poni file'

    def run(self):
        fpath = self.inputs['file_path']
        for line in open(fpath,'r'):
            kv = line.strip().split('=')

            if 'sample_to_CCD_mm' in kv[0]:
                d_mm = float(kv[1])
            if 'pixel_size_x_mm' in kv[0]:
                pxsz_x_mm = float(kv[1])
            if 'pixel_size_y_mm' in kv[0]:
                pxsz_y_mm = float(kv[1])
            if 'beam_center_x_pix' in kv[0]:
                bcx_px = float(kv[1])       # Nika reports the x coord relative to 'bottom left' corner of detector
                                            # where beam axis intersects detector plane, in pixels 
            if 'beam_center_y_pix' in kv[0]:
                bcy_px = float(kv[1])       # same as beam_center_x_pix but for y 
            if 'horizontal_tilt_deg' in kv[0]:    
                htilt_deg = float(kv[1])    # Nika reports the horizontal tilt in degrees...
            if 'vertical_tilt_deg' in kv[0]:
                vtilt_deg = float(kv[1])    # Nika reports the vertical tilt in degrees...
            if 'wavelength_A' in kv[0]:
                wl_A = float(kv[1])         # Nika reports wavelength is in Angstroms

        # get wavelength in m 
        wl_m = wl_A*1E-10
        pxsz_x_um = pxsz_x_mm * 1000
        pxsz_y_um = pxsz_y_mm * 1000
        # TODO: check whether these rotation angle mappings are correct. 
        tilt_deg = -1.*htilt_deg
        rot_fit2d = vtilt_deg
        # use a pyFAI.AzimuthalIntegrator() to do the conversion
        p = pyFAI.AzimuthalIntegrator(wavelength = wl_m) 
        p.setFit2D(d_mm,bcx_px,bcy_px,tilt_deg,rot_fit2d,pxsz_x_um,pxsz_y_um)
        poni_dict = p.getPyFAI()
        self.outputs['poni_dict'] = poni_dict 




from __future__ import print_function
from collections import OrderedDict
import os

from pypif import pif
from citrination_client import CitrinationClient as CitCli
from citrination_client import PifSystemReturningQuery, PifSystemQuery, DataQuery, DatasetQuery, IdQuery, FieldQuery, Filter

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    address=None,
    api_key_file=None)

class CitrinationClient(PawsPlugin):
    """PAWS Plugin wrapping a Citrination client"""

    def __init__(self):
        super(CitrinationClient,self).__init__(content)
        self.content_doc['address'] = 'web address of citrination instance'
        self.content_doc['api_key_file'] = 'path to a file in the local filesystem containing a valid citrination api key'
        self.client=None

    def start(self):
        super(CitrinationClient,self).start()
        self.address = self.content['address'] 
        f = open(self.content['api_key_file'],'r')
        self.api_key = str(f.readline()).strip()
        f.close()
        self.client = CitCli(self.api_key,self.address)

    def stop(self):
        self.client = None

    def upload_pif(self,pif_object,dataset_id,json_path,keep_json=True,upload=True):
        """Upload a PIF to Citrination.

        Parameters
        ----------
        pif_object : object
            A pypif.obj.System object or an array/list thereof
        dataset_id : int
            Integer dataset id where the pif(s) will be uploaded 
        json_path : str
            Path on local filesystem where PIF json data will be dumped
        keep_json : bool
            Flag for whether or not to keep the json dump file 
        upload : False
            Flag for whether or not to perform upload- set to False for a dry run

        Returns
        -------
        resp : str
            Response from the Citrination server.
        """
        if not os.path.splitext(json_path)[1] in ['.json','.JSON']:
            json_path = json_path+'.json'
        jsfnm = os.path.split(json_path)[1]
        self.message_callback('PIF dump file: {}'.format(json_path))
        pif.dump(pif_object, open(json_path,'w'))
        if upload:
            self.message_callback('Uploading {} to dataset {}'.format(json_path,dataset_id))
            try:
                resp = self.client.upload(dataset_id,json_path,jsfnm)
            except:
                resp = 'An error occurred during upload- aborting'
                self.message_callback(resp)
                raise
        else:
            resp = 'upload flag is set to False- no upload occurred'
            self.message_callback(resp)
        if not keep_json:
            self.message_callback('Removing {}'.format(json_path))
            os.remove(json_path) 
        return resp

    def download_pifs(dataset_id,experiment_id=None):
        """Download PIFs from Citrination.

        Parameters
        ----------
        dataset_id : int
            Integer dataset id from where the pif(s) will be downloaded 
        experiment_id : str 
            Optional string for filtering pifs on their experiment_id attributes 
        
        Returns
        -------
        pifs : object
            One or more pypif.obj.System objects
        """
        if experiment_id is not None:
            query = self.dsid_query_with_expt_id(dataset_id,experiment_id)
        else:
            query = self.dsid_query(dataset_id)
        all_hits = []
        n_hits = 0
        self.message_callback('Querying Citrination for records.')
        current_result = self.client.search(query)
        while current_result.hits is not None:
            all_hits.extend(current_result.hits)
            n_current_hits = len(current_result.hits)
            n_hits += n_current_hits
            query.from_index += n_current_hits 
            current_result = self.client.search(query)
        self.message_callback('Found {} records.'.format(n_hits))
        pifs = [x.system for x in all_hits]
        return pifs 



from collections import OrderedDict

from xrsdkit.tools.ymltools import save_sys_to_yaml

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    system={}
    )
outputs=OrderedDict()

class SaveXRSDSystem(Operation):
    """Save xrsdkit.system.System object to a YAML file."""

    def __init__(self):
        super(SaveXRSDSystem, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path where YAML file will be saved'
        self.input_doc['system'] = 'xrsdkit.system.System object'

    def run(self):
        self.message_callback('writing to {}'.format(self.inputs['file_path']))
        save_sys_to_yaml(self.inputs['file_path'],self.inputs['system'])




from collections import OrderedDict
import os

import numpy as np

from ..Operation import Operation

inputs=OrderedDict(
    data=None,
    header='',
    file_path=None,
    delimiter=' ')

outputs=OrderedDict(
    dir_path=None,
    filename=None)

class NumpySave(Operation):
    """Wrapper for numpy.savetxt."""

    def __init__(self):
        super(NumpySave, self).__init__(inputs, outputs)
        self.input_doc['data'] = 'data to save to file'
        self.input_doc['header'] = 'header (string) for the output file'
        self.input_doc['file_path'] = 'path where file will be written'
        self.output_doc['dir_path'] = 'directory of the output file' 
        self.output_doc['filename'] = 'name of the output file' 

    def run(self):
        p = self.inputs['file_path']
        self.message_callback('saving to {}'.format(p))
        dlm = self.inputs['delimiter']
        self.outputs['dir_path'],self.outputs['filename'] = os.path.split(p)
        h = self.inputs['header']
        np.savetxt(p, self.inputs['data'], delimiter=dlm, newline=os.linesep, header=h)




import os
import copy
from collections import OrderedDict

from ...Workflow import Workflow 
from . import Read
from ....operations.IO.FILESYSTEM.BuildFileList import BuildFileList

inputs = OrderedDict(
    header_dir = '',
    header_regex = '*.yml',
    header_suffix = '',
    #header_filter_regex = '.*',
    time_key = 't_utc',
    image_dir = '',
    image_suffix = '',
    image_ext = '.tif',
    q_I_dir = '',
    q_I_suffix = '',
    q_I_ext = '.dat',
    system_dir = '',
    system_suffix = '',
    system_ext = '.yml'
    )

outputs = copy.deepcopy(Read.outputs)
for k in Read.outputs.keys(): outputs[k] = []
outputs.update(
    filenames = [],
    header_files = [],
    image_files = [],
    q_I_files = [],
    system_files = []
    )
    

class ReadBatch(Workflow):

    def __init__(self):
        super(ReadBatch,self).__init__(inputs,outputs)
        self.add_operation('header_files',BuildFileList())
        self.add_operation('read',Read.Read())

    def run(self):
        # initialize outputs in case of Workflow re-use!
        self.outputs = copy.deepcopy(outputs)
        self.operations['header_files'].run_with(
            dir_path = self.inputs['header_dir'],
            regex = self.inputs['header_regex']
            )
        header_file_list = self.operations['header_files'].outputs['file_list']
        self.outputs['header_files'] = header_file_list
        filename_list = [os.path.splitext(os.path.split(hf)[1])[0] for hf in header_file_list]
        hdr_fn_sfx = self.inputs['header_suffix']
        if hdr_fn_sfx: filename_list = [fn[:fn.rfind(hdr_fn_sfx)] for fn in filename_list]

        self.outputs['filenames'] = filename_list 
        q_I_dir = self.inputs['q_I_dir']
        q_I_suffix = self.inputs['q_I_suffix']
        q_I_ext = self.inputs['q_I_ext']
        img_dir = self.inputs['image_dir']
        img_sfx = self.inputs['image_suffix']
        img_ext = self.inputs['image_ext']
        sys_dir = self.inputs['system_dir']
        sys_suffix = self.inputs['system_suffix']
        sys_ext = self.inputs['system_ext']
        if not img_ext[0] == '.': img_ext='.'+img_ext
        if not q_I_ext[0] == '.': q_I_ext='.'+q_I_ext
        if not sys_ext[0] == '.': sys_ext='.'+sys_ext
        image_file_list = [None for fn in filename_list]
        if img_dir and img_ext:
            image_file_list = [os.path.join(img_dir,fn+img_sfx+img_ext) for fn in filename_list]
        self.outputs['image_files'] = image_file_list
        q_I_file_list = [None for fn in filename_list]
        if q_I_dir and q_I_ext:
            q_I_file_list = [os.path.join(q_I_dir,fn+q_I_suffix+q_I_ext) for fn in filename_list]
        self.outputs['q_I_files'] = q_I_file_list
        system_file_list = [None for fn in filename_list]
        if sys_dir and sys_ext:
            system_file_list = [os.path.join(sys_dir,fn+sys_suffix+sys_ext) for fn in filename_list]
        self.outputs['system_files'] = system_file_list

        n_hdrs = len(filename_list)
        self.message_callback('STARTING BATCH ({})'.format(n_hdrs))
        for ihdr, hdr_fn, img_fn, q_I_fn, sys_fn in zip(range(n_hdrs),
        header_file_list,image_file_list,q_I_file_list,system_file_list):
            self.message_callback('RUNNING {} / {}'.format(ihdr+1,n_hdrs))
            outs = self.operations['read'].run_with(
                header_file = hdr_fn,
                time_key = self.inputs['time_key'],
                image_file = img_fn,
                q_I_file = q_I_fn,
                system_file = sys_fn
                )        
            for out_key, out_data in outs.items():
                self.outputs[out_key].append(out_data)
        return self.outputs




import numpy as np
from collections import OrderedDict

from ...Operation import Operation

inputs = OrderedDict(
    q_I_meas=None,
    T_meas=None,
    bg_batch_outputs=None,
    q_I_bg_key=None,
    T_bg_key=None)
outputs = OrderedDict(
    q_I_bgsub=None,
    T_bg=None,
    bg_factor=None)

class BgSubtractByTemperature(Operation):
    """
    Find a background spectrum from a batch of background spectra,
    where the temperature of the background spectrum is as close as possible
    to the (input) temperature of the measured spectrum.
    Then subtract that background spectrum from the input spectrum.
    The measured and background spectra are expected to have the same domain.

    Operation originally contributed by Amanda Fournier.  
    """
    
    def __init__(self):
        super(BgSubtractByTemperature, self).__init__(inputs, outputs)
        self.input_doc['q_I_meas'] = 'n-by-2 array of I(q) versus q'
        self.input_doc['T_meas'] = 'temperature as taken from the dict '\
            'produced by the detector header file'
        self.input_doc['bg_batch_outputs'] = 'the output (list of dicts) '\
            'of a batch of background spectra at different temperatures'
        self.input_doc['q_I_bg_key'] = 'the name of the bg_batch workflow output '\
            'containing n-by-2 array of q and I(q) for the background spectra '
        self.input_doc['T_bg_key'] = 'the name of the bg_batch workflow output '\
            'containing the background spectrum temperatures'
        self.output_doc['q_I_bgsub'] = 'n-by-2 array of q and '\
            'background-subracted intensity (I_meas-bg_factor*I_bg)'
        self.output_doc['T_bg'] = 'temperature of the subtracted background spectrum'
        self.output_doc['bg_factor'] = 'correction factor applied to background '\
            'before subtraction, to ensure positive background-subtracted intensities'

    def run(self):
        q_I_meas = self.inputs['q_I_meas']
        T_meas = self.inputs['T_meas']
        qIkey = self.inputs['q_I_bg_key']
        Tkey = self.inputs['T_bg_key']
        bg_out = self.inputs['bg_batch_outputs']
        T_allbg = [d[Tkey] for d in bg_out]
        closest_T_idx = np.argmin(np.abs([T_bg - T_meas for T_bg in T_allbg]))
        T_bg = T_allbg[closest_T_idx]
        q_I_bg = bg_out[closest_T_idx][qIkey]
        if not all(q_I_meas[:,0] == q_I_bg[:,0]):
            msg = 'SPECTRUM AND BACKGROUND ON DIFFERENT q DOMAINS'
            raise ValueError(msg)
        bad_data = ( (q_I_meas[:,1] <= 0) 
            | (q_I_bg[:,1] <= 0) 
            | np.isnan(q_I_meas[:,1]) 
            | np.isnan(q_I_bg[:,1]) )
        I_floor = 0 
        #I_floor = 1E-6 * np.max(q_I_meas[:,1])
        bg_factor = np.min((q_I_meas[:,1][~bad_data]-I_floor) / q_I_bg[:,1][~bad_data])
        q_I_bgsub = np.array(q_I_meas)
        q_I_bgsub[:,1] = q_I_meas[:,1] - bg_factor * q_I_bg[:,1]

        self.outputs['q_I_bgsub'] = q_I_bgsub
        self.outputs['T_bg'] = T_bg 
        self.outputs['bg_factor'] = bg_factor





from collections import OrderedDict
import os
import time

from ...Operation import Operation

inputs=OrderedDict(ppumps=None)
outputs=OrderedDict()
        
class StopPPumps(Operation):
    """Set an array of MitosPPumpControllers to idle."""

    def __init__(self):
        super(StopPPumps,self).__init__(inputs,outputs)
        self.input_doc['ppumps'] = 'dict of MitosPPumpController plugins'

    def run(self):
        ppcs = self.inputs['ppumps'] 
        for pp_nm,pp in ppcs.items():
            pp.set_idle()
            #ppc.set_flowrate(0)




from collections import OrderedDict
import os

import numpy as np
import yaml

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    output_keymap={})
outputs=OrderedDict(
    data=None,
    dir_path=None,
    filename=None)

class LoadYAML(Operation):
    """Load data from a YAML file with yaml.load()"""

    def __init__(self):
        super(LoadYAML, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to YAML-formatted data file'
        self.input_doc['output_keymap'] = 'dict of key-map pairs: '\
            'each key is added to the output `data` dict, '\
            'and each map is a list of keys '\
            'for extracting an embedded value from `data` '\
            '(i.e., for k,v in output_keymap.items(): data[k] = data[v[0]][v[1]][...])'
        self.output_doc['data'] = 'the result of yaml.load(file_path)' 
        self.output_doc['dir_path'] = 'directory path'
        self.output_doc['filename'] = 'filename with path and extension stripped'

    def run(self):
        p = self.inputs['file_path']
        self.message_callback('loading {}'.format(p))
        dir_path,filename = os.path.split(p)
        file_noext = os.path.splitext(filename)[0]
        self.outputs['dir_path'] = dir_path 
        self.outputs['filename'] = file_noext 
        f = open(p,'r')
        ds = yaml.load(f)
        f.close()

        for k,ks in self.inputs['output_keymap'].items():
            v = ds[ks[0]] 
            for kk in ks[1:]:
                v = v[kk]
            ds[k] = v
        self.outputs['data'] = ds 

        return self.outputs




from collections import OrderedDict

from ...Operation import Operation

inputs=OrderedDict(
    ssh_client=None,
    remote_path=None,
    local_path=None)
outputs=OrderedDict()

class SSHCopy(Operation):
    """Use SSHClient Plugin to get a file from remote filesystem"""

    def __init__(self):
        super(SSHCopy, self).__init__(inputs, outputs)
        self.input_doc['ssh_client'] = 'SSHClient plugin'
        self.input_doc['remote_path'] = 'remote file path (file source)'
        self.input_doc['local_path'] = 'local file path (file destination)'

    def run(self):
        rp = self.inputs['remote_path']
        lp = self.inputs['local_path']
        self.inputs['ssh_client'].copy_file(rp,lp)
        #self.inputs['ssh_client'].request_copy(rdir,rx,ldir,blk)




from __future__ import print_function
from collections import OrderedDict
import importlib

from ..pawstools.DictTree import DictTree 
from .. import operations as ops
from ..pawstools import OperationLoadError

class OpManager(DictTree):
    """OpManager provides access to and control over paws Operations."""

    def __init__(self):
        # TODO: handle flags
        #default_flags = OrderedDict()
        #default_flags['enabled'] = False
        super(OpManager,self).__init__()
        self.ops_enabled = DictTree()
        self.message_callback = self.tagged_print 
        self.cat_op_list = []
        self.load_operations()

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def get_operation(self,op_key):
        """Get an Operation, enable it if needed, instantiate, return.

        Parameters
        ----------
        op_key : str
            key indicating the operation module,
            e.g. PROCESSING.TESTS.Fibonacci.

        Returns
        -------
        op : Operation
            the instantiated Operation 
        """
        if not self.is_op_enabled(op_key):
            try:
                self.enable_op(op_key)
            except Exception:
                raise OperationLoadError('Error loading {}'.format(op_key)) 
        op = self.get_data(op_key)
        return op()

    def load_operations(self,cat_op_list=ops.cat_op_list):
        """Load Operations into OpManager.

        Parameters
        ----------
        cat_op_list : list of (str,str), optional
            Each element in the list is a tuple containing two strings.
            The first string in the tuple indicates the category subpackage.
            The second string in the tuple indicates the operation module.
            The default is to load all operations 
            detected by paws.operations at startup.
        """
        for cat_op in cat_op_list:
            self.add_cat(cat_op[0])
            if cat_op[1]:
                self.add_op(cat_op[0],cat_op[1])
                self.cat_op_list.append((cat_op[0],cat_op[1]))

    def add_cat(self,cat_module):
        """Add category `cat_module` to the tree."""
        cat_tags = cat_module.split('.')
        cat_key = cat_tags[0]
        if not cat_key in self.keys():
            self.set_data(cat_key,{})
            self.ops_enabled.set_data(cat_key,{})
        if len(cat_tags) > 1:
            for cat_tag in cat_tags[1:]:
                cat_key = cat_key+'.'+cat_tag
                if not cat_key in self.keys():
                    self.set_data(cat_key,{})
                    self.ops_enabled.set_data(cat_key,{})

    def add_op(self,cat_module,op_name):
        """Add label for an Operation at `op_name` under category `cat_module`."""
        op_key = cat_module+'.'+op_name
        self.set_data(op_key,None)
        self.ops_enabled.set_data(op_key,False)

    def enable_op(self,op_module):
        """Import Operation module and add its Operation to the tree.

        This method imports the Operation to check compatibility,
        and then sets the 'enabled' flag to True.
        After this, the Operation is available via self.get_op()

        Parameters
        ----------
        op_module : str
            Name of the Operation module.
            Example: If class MyOperation is in the CATEGORY.MyOperation module,
            retrieve it with `op_module` = 'CATEGORY.MyOperation'.
        """
        op_name = op_module.split('.')[-1]
        try:
            m = importlib.import_module('.'+op_module,ops.__name__)
        except ImportError:
            raise OperationLoadError('Error loading {}'.format(op_module)) 
        op = getattr(m,op_name)
        optest = op()
        self.set_data(op_module,op)
        self.ops_enabled[op_module] = True

    def is_op_enabled(self,op_module):
        """Return boolean indicating whether Operation is enabled.

        Parameters
        ----------
        op_module : str
            Name of the Operation module.
            see enable_op().
        """
        return self.ops_enabled[op_module] 

    def n_ops(self):
        return len(self.cat_op_list) 

    def list_operations(self):
        return [catnm+'.'+opnm for catnm,opnm in self.cat_op_list] 





from collections import OrderedDict
import time
import os

from ...Operation import Operation

inputs=OrderedDict(cryocon=None)
outputs=OrderedDict()
        
class StopCryoCon(Operation):
    """Stop control of the CryoCon"""

    def __init__(self):
        super(StopCryoCon,self).__init__(inputs,outputs)
        self.input_doc['cryocon'] = 'A CryoConController Plugin'

    def run(self):
        self.inputs['cryocon'].stop_control()




from collections import OrderedDict
import os

import numpy as np
import pyFAI

from ...Operation import Operation

inputs=OrderedDict(
    fit2d_file=None,
    wavelength_A=None)
outputs=OrderedDict(
    poni_dict=None,
    fit2d_dict=None)

class Fit2DToPONI(Operation):
    """Produce PONI calibration parameters from a Fit2D calibration result. 

    Reads in Fit2D calibration result from a text file,
    converts it to a dict of PyFAI PONI parameters.

    WARNING: Some developers have claimed that 
    the Fit2D software itself computes tilted geometries incorrectly.
    If using actual Fit2D output with tilts, 
    it is suggested that a new calibration be attempted,
    with a different package (something other than Fit2D).
    Use tilts from Fit2D output at your own risk.
 
    Format of text file for Fit2D parameters is expected to be:
    directDist=____         # sample-detector
    centerX=____            # beam center on detector, in pixels
    centerY=____            # beam center on detector, in pixels
    pixelX=____             # pixel width, microns
    pixelY=____             # pixel height, microns
    tilt=____               # degrees
    tiltPlanRotation=____   # degrees
    splineFile=____         # optional detector distortion correction file
    """
    
    def __init__(self):
        super(Fit2DToPONI,self).__init__(inputs,outputs)
        self.input_doc['fit2d_file'] = 'text file describing Fit2D geometry'
        self.input_doc['wavelength_A'] = 'wavelength in Angstroms'
        self.output_doc['fit2d_dict'] = 'Fit2D calibration parameters'
        self.output_doc['poni_dict'] = 'pyFAI calibration parameters'

    def run(self):
        fpath = self.inputs['fit2d_file']
        fit2d_dict = OrderedDict() 
        for line in open(fpath,'r'):
            kv = line.strip().split('=')
            fit2d_dict[kv[0]] = float(kv[1])
        # get wavelength in m 
        wl_m = self.inputs['wavelength_A']*1E-10
        # use a pyFAI.AzimuthalIntegrator() to do the conversion
        p = pyFAI.AzimuthalIntegrator(wavelength = wl_m) 
        p.setFit2D(
            fit2d_dict['directDist'],
            fit2d_dict['centerX'],
            fit2d_dict['centerY'],
            fit2d_dict['tilt'],
            fit2d_dict['tiltPlanRotation'],
            fit2d_dict['pixelX'],
            fit2d_dict['pixelY'])
        poni_dict = p.getPyFAI()
        self.outputs['fit2d_dict'] = fit2d_dict 
        self.outputs['poni_dict'] = poni_dict 




from __future__ import print_function
from collections import OrderedDict
import os
import copy
import time

from citrination_client.models.design import Target,constraints 

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    citrination_client=None,
    dataview_id=None,
    target=None,
    target_goal='Min',
    constraints={},
    range_constraints={},
    categorical_constraints={},
    n_candidates=2,
    design_effort=2)

class FlowDesigner(PawsPlugin):

    def __init__(self):
        super(FlowDesigner,self).__init__(content)
        self.thread_blocking = False
        self.content_doc['citrination_client'] = 'A running CitrinationClient plugin'
        self.content_doc['dataview_id'] = 'integer id of the data view to query' 
        self.content_doc['target'] = 'property name to optimize' 
        self.content_doc['target_goal'] = 'Min or Max, depending on target optimization goals' 
        self.content_doc['constraints'] = 'dict of property names and real constraint values' 
        self.content_doc['range_constraints'] = 'dict of property names and list of values ([min,max])' 
        self.content_doc['categorical_constraints'] = 'dict of property names and categorical constraint values' 
        self.content_doc['n_candidates'] = 'number of design candidates to request per iteration' 
        self.content_doc['design_effort'] = 'how hard to try to meet the targets (int from 1 to 30)' 
        self.best_materials = []
        self.next_experiments = []

    def description(self):
        desc = 'FlowDesigner Plugin: '\
            'Queries a data view on Citrination '\
            'to predict the flow rates needed for '\
            'a FlowReactor Plugin to achieve a target.'
        return desc

    def start(self):
        super(FlowDesigner,self).start() 

    def set_design_goal(self,design_goal):
        self.content['target']=design_goal['target']
        self.content['target_goal']=design_goal['target_goal']
        self.content['constraints']=design_goal['constraints']
        self.content['range_constraints']=design_goal['range_constraints']
        self.content['categorical_constraints']=design_goal['categorical_constraints']

    def get_candidate_recipes(self):
        cc = self.content['citrination_client']
        dvid = self.content['dataview_id']
        tgt = Target(self.content['target'],self.content['target_goal'])
        straints = []
        n_candidates = self.content['n_candidates'] 
        design_effort = self.content['design_effort'] 
        for prop_name, val in self.content['constraints'].items():
            straints.append(constraints.RealValueConstraint(prop_name,val))
        for prop_name, lmts in self.content['range_constraints'].items():
            straints.append(constraints.RealRangeConstraint(prop_name,lmts[0],lmts[1]))
        for prop_name, cats in self.content['categorical_constraints'].items():
            straints.append(constraints.CategoricalConstraint(prop_name,cats))
        #DOC: cc.submit_design_run(
        #       data_view_id,
        #       num_candidates (int in [1,20]),
        #       effort (int in [1,30]),
        #       target=None, constraints=[],
        #       sampler='Default') 
        msg = 'Designing for: \nTarget: {} \nConstraints: {} \nRange constraints: {} \nCategorical constraints: {}'.format(
            self.content['target'],self.content['constraints'],
            self.content['range_constraints'],self.content['categorical_constraints']
            )
        self.message_callback(msg)
        #try_again = True
        #while try_again:
        #    try:
        des = cc.client.submit_design_run(dvid,n_candidates,design_effort,tgt,straints) 
        fin = False
        while not fin:
            time.sleep(2)
            stat = cc.client.get_design_run_status(dvid, des.uuid)
            self.message_callback('design finished: {} ({}/100)'.format(stat.finished(),stat.progress))
            if int(stat.progress) == 100:
                fin = True
        desres = cc.client.get_design_run_results(dvid,des.uuid)
        #        try_again = False 
        #    except:
        #        pass
        self.best_materials = desres.best_materials
        self.next_experiments = desres.next_experiments




import os

from paws import pawstools
from paws.workflows.WfManager import WfManager

wfmgr = WfManager()

wfmgr.load_packaged_wfm('FITTING.BL15.LEGACY.batch_fit')

wf = wfmgr.workflows['batch_fit']

wf.set_op_inputs('t_filenames',x_sort_flag=True)
wf.set_op_inputs('t_q_I_files',x_sort_flag=True)
wf.set_op_inputs('t_system_files',x_sort_flag=True)
wf.set_op_input('run_fits','serial_params',{'system':'system'})

wfmgr.save_to_wfm(os.path.join(pawstools.sourcedir,'workflows','FITTING','BL15','LEGACY','timeseries_fit.wfm'))




import os
import copy
from collections import OrderedDict

from paws.workflows.Workflow import Workflow 
from paws.workflows.INTEGRATION import Integrate 
from paws.operations.IO.FILESYSTEM.BuildFileList import BuildFileList

inputs = OrderedDict(
    image_data = [],
    image_files = [],
    image_dir = None,
    image_regex = None,
    integrator = None,
    q_min = 0.,
    q_max = float('inf'),
    output_dir = None,
    output_filenames = []
    )

outputs = copy.deepcopy(Integrate.outputs)
for k in Integrate.outputs.keys(): outputs[k] = []

class IntegrateBatch(Workflow):

    def __init__(self):
        super(IntegrateBatch,self).__init__(inputs,outputs)
        self.add_operations(
            image_files=BuildFileList(),
            integrate=Integrate.Integrate()
            )

    def run(self):
        self.outputs = copy.deepcopy(outputs)
        img_files = self.inputs['image_files']
        img_data = self.inputs['image_data']
        out_fns = self.inputs['output_filenames']
        if self.inputs['image_dir'] and self.inputs['image_regex']:
            file_outputs = self.operations['image_files'].run_with(
                dir_path = self.inputs['image_dir'],
                regex = self.inputs['image_regex']
                )
            img_files = file_outputs['file_list']
        self.operations['integrate'].set_inputs(
            integrator = self.inputs['integrator'],
            q_min = self.inputs['q_min'],
            q_max = self.inputs['q_max'],
            output_dir = self.inputs['output_dir']
            )
        if not img_data:
            # assume img_files were provided
            img_data = [None for fn in img_files]
        elif not img_files:
            # assume img_data were provided
            img_files = [None for dd in img_data]
        if not out_fns:
            out_fns = [None for fn in img_files]  
        for fn,fd,outfn in zip(img_files,img_data,out_fns):
            if not outfn:
                outfn = os.path.splitext(os.path.split(fn)[1])[0]
            integ_outputs = self.operations['integrate'].run_with(
                image_file = fn,
                image_data = fd,
                output_filename = outfn 
                )
            for out_key, out_data in integ_outputs.items():
                self.outputs[out_key].append(out_data)
        return self.outputs




import os
import copy
from collections import OrderedDict

from paws.workflows.Workflow import Workflow 

from paws.workflows.IO.BL15 import ReadTimeSeries
from paws.workflows.FITTING import XRSDFit

inputs = copy.deepcopy(ReadTimeSeries.inputs)
inputs.update(copy.deepcopy(XRSDFit.inputs))
inputs.update({'experiment_id':None})

outputs = copy.deepcopy(ReadTimeSeries.outputs)
outputs.update(copy.deepcopy(XRSDFit.outputs))
for k in XRSDFit.outputs.keys(): outputs[k] = []

class TimeSeriesXRSDFit(Workflow):

    def __init__(self):
        super(TimeSeriesXRSDFit,self).__init__(inputs,outputs)
        self.add_operation('read_timeseries',ReadTimeSeries.ReadTimeSeries())
        self.add_operation('fit',XRSDFit.XRSDFit())

    def run(self):
        read_inputs = OrderedDict([(k,self.inputs[k]) for k in ReadTimeSeries.inputs.keys()])
        self.operations['read_timeseries'].operations['read_batch'].operations['read'].disable_op('read_image')
        ts_data = self.operations['read_timeseries'].run_with(**read_inputs) 
        nb = len(ts_data['system_files'])
        self.message_callback('STARTING BATCH ({})'.format(nb))
        previous_sys = None
        for ib,outfile,q_I,sys,hdr,q_I_file in zip(range(nb),
        ts_data['system_files'],ts_data['q_I'],ts_data['system'],
        ts_data['header_data'],ts_data['q_I_files']):
            self.message_callback('RUNNING {} / {}'.format(ib+1,nb))
            if not sys:
                if previous_sys:
                    sys = previous_sys
                else:
                    sys = self.inputs['system']
            samp_id = None
            if self.inputs['experiment_id'] and hdr:
                if 't_utc' in hdr:
                    samp_id = str(self.inputs['experiment_id'])+'_'+str(int(hdr['t_utc']))
            data_fp = None
            if q_I_file: data_fp = os.path.split(q_I_file)[1]
            fit_outputs = self.operations['fit'].run_with(
                system = sys,
                q_I = q_I,
                error_weighted = self.inputs['error_weighted'],
                logI_weighted = self.inputs['logI_weighted'],
                q_range = self.inputs['q_range'],
                output_file = outfile,
                experiment_id = self.inputs['experiment_id'],
                sample_id = samp_id, 
                data_file_path = data_fp 
                )
            previous_sys = fit_outputs['system_opt']
        return self.outputs
